# -*- coding: utf-8 -*-

from collective.contact.plonegroup.utils import get_organizations
from collective.eeafaceted.batchactions.interfaces import IBatchActionsMarker
from copy import deepcopy
from eea.facetednavigation.criteria.interfaces import ICriteria
from imio.helpers.catalog import addOrUpdateColumns
from imio.helpers.catalog import addOrUpdateIndexes
from imio.helpers.content import get_vocab
from imio.helpers.content import get_vocab_values
from imio.helpers.content import richtextval
from imio.helpers.content import safe_delattr
from imio.pyutils.utils import replace_in_list
from OFS.ObjectManager import BeforeDeleteException
from persistent.mapping import PersistentMapping
from plone import api
from plone.app.contenttypes.migration.migration import migrate as pac_migrate
from Products.Archetypes.event import ObjectEditedEvent
from Products.CMFPlone.utils import base_hasattr
from Products.CMFPlone.utils import safe_unicode
from Products.contentmigration.basemigrator.migrator import CMFFolderMigrator
from Products.cron4plone.browser.configlets.cron_configuration import ICronConfiguration
from Products.GenericSetup.tool import DEPENDENCY_STRATEGY_NEW
from Products.PloneMeeting.browser.itemattendee import position_type_default
from Products.PloneMeeting.config import AddAdvice
from Products.PloneMeeting.content.advice import IMeetingAdvice
from Products.PloneMeeting.content.meeting import IMeeting
from Products.PloneMeeting.events import onHeldPositionWillBeRemoved
from Products.PloneMeeting.interfaces import IMeetingDashboardBatchActionsMarker
from Products.PloneMeeting.interfaces import IMeetingItemDashboardBatchActionsMarker
from Products.PloneMeeting.MeetingConfig import PROPOSINGGROUPPREFIX
from Products.PloneMeeting.migrations import logger
from Products.PloneMeeting.migrations import Migrator
from Products.PloneMeeting.migrations.migrate_to_4209 import Migrate_To_4209
from Products.PloneMeeting.profiles import MeetingConfigDescriptor
from Products.PloneMeeting.setuphandlers import columnInfos
from Products.PloneMeeting.setuphandlers import indexInfos
from Products.PloneMeeting.utils import cleanMemoize
from Products.ZCatalog.ProgressHandler import ZLogHandler
from zope.annotation import IAnnotations
from zope.component import queryUtility
from zope.event import notify
from zope.interface import alsoProvides
from zope.interface import noLongerProvides


class MeetingMigrator(CMFFolderMigrator):
    """ """
    src_portal_type = None
    src_meta_type = 'Meeting'
    dst_portal_type = None
    dst_meta_type = None  # not used
    # will store the Migrate_To_4200 migrator instance
    # so we can use it to manage custom migration
    pm_migrator = None

    def migrate_atctmetadata(self):
        """Override to not migrate exclude_from_nav because it does not exist by default
           and it takes parent's value that is an instancemethod and fails at transaction commit..."""
        pass

    def migrate_schema_fields(self):
        # fields
        date = self.old.getDate()
        if date:
            date._timezone_naive = True
            self.new.date = date.asdatetime()
        start_date = self.old.getStartDate()
        if start_date:
            start_date._timezone_naive = True
            self.new.start_date = start_date.asdatetime()
        mid_date = self.old.getMidDate()
        if mid_date:
            mid_date._timezone_naive = True
            self.new.mid_date = mid_date.asdatetime()
        end_date = self.old.getEndDate()
        if end_date:
            end_date._timezone_naive = True
            self.new.end_date = end_date.asdatetime()
        approval_date = self.old.getApprovalDate()
        if approval_date:
            approval_date._timezone_naive = True
            self.new.approval_date = approval_date.asdatetime()
        convocation_date = self.old.getConvocationDate()
        if convocation_date:
            convocation_date._timezone_naive = True
            self.new.convocation_date = convocation_date.asdatetime()
        deadline_publish = self.old.getDeadlinePublish()
        if deadline_publish:
            deadline_publish._timezone_naive = True
            self.new.validation_deadline = deadline_publish.asdatetime()
        deadline_freeze = self.old.getDeadlineFreeze()
        if deadline_freeze:
            deadline_freeze._timezone_naive = True
            self.new.freeze_deadline = deadline_freeze.asdatetime()
        self.new.assembly = self.old.getRawAssembly() and \
            richtextval(self.old.getRawAssembly()) or None
        self.new.assembly_excused = self.old.getRawAssemblyExcused() and \
            richtextval(self.old.getRawAssemblyExcused()) or None
        self.new.assembly_absents = self.old.getRawAssemblyAbsents() and \
            richtextval(self.old.getRawAssemblyAbsents()) or None
        self.new.assembly_guests = self.old.getRawAssemblyGuests() and \
            richtextval(self.old.getRawAssemblyGuests()) or None
        self.new.assembly_proxies = self.old.getRawAssemblyProxies() and \
            richtextval(self.old.getRawAssemblyProxies()) or None
        self.new.assembly_staves = self.old.getRawAssemblyStaves() and \
            richtextval(self.old.getRawAssemblyStaves()) or None
        self.new.signatures = self.old.getRawSignatures() and \
            richtextval(self.old.getRawSignatures()) or None
        # place is moved to place/place_other
        if 'place' in self.used_meeting_attrs:
            place = safe_unicode(self.old.getPlace().strip())
            vocab = get_vocab(
                self.new, "Products.PloneMeeting.content.meeting.places_vocabulary")
            if not place or place not in vocab:
                self.new.place = u'other'
                self.new.place_other = place or None
            else:
                self.new.place = place
        self.new.pre_meeting_place = safe_unicode(self.old.getPreMeetingPlace())
        pre_meeting_date = self.old.getPreMeetingDate()
        if pre_meeting_date:
            pre_meeting_date._timezone_naive = True
            self.new.pre_meeting_date = pre_meeting_date.asdatetime()
        self.new.extraordinary_session = self.old.getExtraordinarySession()
        self.new.in_and_out_moves = self.old.getRawInAndOutMoves() and \
            richtextval(self.old.getRawInAndOutMoves()) or None
        self.new.notes = self.old.getRawNotes() and richtextval(self.old.getRawNotes()) or None
        self.new.observations = self.old.getRawObservations() and \
            richtextval(self.old.getRawObservations()) or None
        self.new.pre_observations = self.old.getRawPreObservations() and \
            richtextval(self.old.getRawPreObservations()) or None
        self.new.committees_observations = self.old.getRawCommitteeObservations() and \
            richtextval(self.old.getRawCommitteeObservations()) or None
        self.new.votes_observations = self.old.getRawVotesObservations() and \
            richtextval(self.old.getRawVotesObservations()) or None
        self.new.public_meeting_observations = self.old.getRawPublicMeetingObservations() and \
            richtextval(self.old.getRawPublicMeetingObservations()) or None
        self.new.secret_meeting_observations = self.old.getRawSecretMeetingObservations() and \
            richtextval(self.old.getRawSecretMeetingObservations()) or None
        self.new.authority_notice = self.old.getRawAuthorityNotice() and \
            richtextval(self.old.getRawAuthorityNotice()) or None
        self.new.meeting_number = self.old.getMeetingNumber()
        self.new.first_item_number = self.old.getFirstItemNumber()
        # custom attributes
        self.new.item_absents = deepcopy(self.old.itemAbsents)
        self.new.item_excused = deepcopy(self.old.itemExcused)
        self.new.item_non_attendees = deepcopy(self.old.itemNonAttendees)
        self.new.item_signatories = deepcopy(self.old.itemSignatories)
        self.new.ordered_contacts = deepcopy(self.old.orderedContacts)

    def migrate(self):
        """ """
        super(MeetingMigrator, self).migrate()
        self.new.update_title()
        self.pm_migrator._hook_custom_meeting_to_dx(self.old, self.new)
        # we use idxs=() because when passing nothing (so idxs=[])
        # then notifyModified is called and the element is modified
        self.new.reindexObject(idxs=())


class Migrate_To_4200(Migrator):

    def _migrateMeetingToDX(self):
        '''Migrate from AT MeetingCategory to DX meetingcategory.'''
        logger.info('Migrating Meeting from AT to DX...')
        # prepare migration to DX
        # manage link between item and meeting manually
        self._removeMeetingItemsReferenceField()
        # update stored Meeting.itemSignatories
        self._updateItemSignatories()
        # configure votes
        self._configureVotes()
        # unregister Meeting portal_types from portal_factory
        portal_factory = self.portal.portal_factory
        registered_types = [portal_type for portal_type in portal_factory.getFactoryTypes().keys()
                            if not portal_type.startswith('Meeting') or
                            portal_type.startswith('MeetingItem')]
        portal_factory.manage_setPortalFactoryTypes(listOfTypeIds=registered_types)
        # main migrate meetings to DX
        self.request.set('currently_migrating_meeting_dx', True)
        for cfg in self.tool.objectValues("MeetingConfig"):
            # update MeetingConfig attributes
            # usedMeetingAttributes
            used_attrs = cfg.getUsedMeetingAttributes()
            used_attrs = replace_in_list(used_attrs, "startDate", "start_date")
            used_attrs = replace_in_list(used_attrs, "midDate", "mid_date")
            used_attrs = replace_in_list(used_attrs, "endDate", "end_date")
            used_attrs = replace_in_list(used_attrs, "approvalDate", "approval_date")
            used_attrs = replace_in_list(used_attrs, "convocationDate", "convocation_date")
            used_attrs = replace_in_list(used_attrs, "assemblyExcused", "assembly_excused")
            used_attrs = replace_in_list(used_attrs, "assemblyAbsents", "assembly_absents")
            used_attrs = replace_in_list(used_attrs, "assemblyGuests", "assembly_guests")
            used_attrs = replace_in_list(used_attrs, "assemblyProxies", "assembly_proxies")
            used_attrs = replace_in_list(used_attrs, "assemblyStaves", "assembly_staves")
            used_attrs = replace_in_list(used_attrs, "extraordinarySession", "extraordinary_session")
            used_attrs = replace_in_list(used_attrs, "inAndOutMoves", "in_and_out_moves")
            used_attrs = replace_in_list(used_attrs, "preMeetingDate", "pre_meeting_date")
            used_attrs = replace_in_list(used_attrs, "preMeetingPlace", "pre_meeting_place")
            used_attrs = replace_in_list(used_attrs, "preObservations", "pre_observations")
            used_attrs = replace_in_list(used_attrs, "committeeObservations", "committees_observations")
            used_attrs = replace_in_list(used_attrs, "votesObservations", "votes_observations")
            used_attrs = replace_in_list(used_attrs, "publicMeetingObservations", "public_meeting_observations")
            used_attrs = replace_in_list(used_attrs, "secretMeetingObservations", "secret_meeting_observations")
            used_attrs = replace_in_list(used_attrs, "authorityNotice", "authority_notice")
            used_attrs = replace_in_list(used_attrs, "meetingNumber", "meeting_number")
            used_attrs = replace_in_list(used_attrs, "firstItemNumber", "first_item_number")
            used_attrs = replace_in_list(used_attrs, "nonAttendees", "non_attendees")
            used_attrs = replace_in_list(used_attrs, "deadlinePublish", "validation_deadline")
            used_attrs = replace_in_list(used_attrs, "deadlineFreeze", "freeze_deadline")
            cfg.setUsedMeetingAttributes(used_attrs)
            # xhtmlTransformFields
            fields = cfg.getXhtmlTransformFields()
            fields = replace_in_list(fields, "Meeting.authorityNotice", "Meeting.authority_notice")
            fields = replace_in_list(fields, "Meeting.committeeObservations", "Meeting.committees_observations")
            fields = replace_in_list(fields, "Meeting.inAndOutMoves", "Meeting.in_and_out_moves")
            fields = replace_in_list(fields, "Meeting.preObservations", "Meeting.pre_observations")
            fields = replace_in_list(fields, "Meeting.publicMeetingObservations", "Meeting.public_meeting_observations")
            fields = replace_in_list(fields, "Meeting.secretMeetingObservations", "Meeting.secret_meeting_observations")
            fields = replace_in_list(fields, "Meeting.votesObservations", "Meeting.votes_observations")
            cfg.setXhtmlTransformFields(fields)
            # meetingColumns
            cols = cfg.getMeetingColumns()
            cols = replace_in_list(cols, "static_startDate", "static_start_date")
            cols = replace_in_list(cols, "static_endDate", "static_end_date")
            cols = replace_in_list(cols, "static_startDate", "static_start_date")
            cfg.setMeetingColumns(cols)

            # remove portal_type so it is created by MeetingConfig.registerPortalTypes
            meeting_type_name = cfg.getMeetingTypeName()
            self.removeUnusedPortalTypes(portal_types=[meeting_type_name])
            cfg.registerPortalTypes()
            MeetingMigrator.src_portal_type = meeting_type_name
            MeetingMigrator.dst_portal_type = meeting_type_name
            MeetingMigrator.used_meeting_attrs = cfg.getUsedMeetingAttributes()
            MeetingMigrator.pm_migrator = self
            pac_migrate(self.portal, MeetingMigrator)

            # some parameters were renamed
            if getattr(cfg, "publishDeadlineDefault", None):
                cfg.setValidationDeadlineDefault(cfg.publishDeadlineDefault)

            # some attributes were removed from MeetingConfig
            safe_delattr(cfg, "historizedMeetingAttributes")
            safe_delattr(cfg, "recordMeetingHistoryStates")
            safe_delattr(cfg, "publishDeadlineDefault")
            safe_delattr(cfg, "preMeetingDateDefault")

        self.request.set('currently_migrating_meeting_dx', False)

        # after migration to DX
        # fix DashboardCollections that use renamed indexes
        self.changeCollectionIndex('getDate', 'meeting_date')
        self.changeCollectionIndex('linkedMeetingUID', 'meeting_uid')
        self.changeCollectionIndex('linkedMeetingDate', 'meeting_date')
        self.changeCollectionIndex('getPreferredMeeting', 'preferred_meeting_uid')
        self.changeCollectionIndex('getPreferredMeetingDate', 'preferred_meeting_date')
        logger.info('Done.')

    def _hook_custom_meeting_to_dx(self, old, new):
        """Hook for plugins that need to do things
           during the Meeting is migrated to DX."""
        pass

    def _hook_before_meeting_to_dx(self):
        """Hook for plugins that need to do things just
           before Meeting is migrated to DX."""
        pass

    def _hook_after_meeting_to_dx(self):
        """Hook for plugins that need to do things just
           after Meeting is migrated to DX."""
        pass

    def _doConfigureItemWFValidationLevels(self, cfg):
        """Method that handles MeetingConfig.itemWFValidationLevels,
           overridable by subplugins if necessary."""
        wfas = cfg.getWorkflowAdaptations()
        stored_itemWFValidationLevels = getattr(cfg, 'itemWFValidationLevels', [])
        # necessary for profile inbetween (4.2beta...)
        if not stored_itemWFValidationLevels:
            itemWFValidationLevels = cfg.getItemWFValidationLevels()
            # a default value exist defining configuration for states
            # itemcreated/proposed/prevalidated/proposedToValidationLevel1/.../proposedToValidationLevel5
            # disable not used states.
            adapted_itemWFValidationLevels = []
            for level in itemWFValidationLevels:
                adapted_level = deepcopy(level)
                # proposedToValidationLevel1-5 are new, disable it by default
                if adapted_level['state'].startswith('proposedToValidationLevel'):
                    adapted_level['enabled'] = '0'
                # itemcreated, enabled by default
                if adapted_level['state'] == 'itemcreated' and 'items_come_validated' in wfas:
                    adapted_level['enabled'] = '0'
                # proposed, enabled by default
                if adapted_level['state'] == 'proposed':
                    if 'items_come_validated' in wfas or 'no_proposal' in wfas:
                        adapted_level['enabled'] = '0'
                    elif ('pre_validation_keep_reviewer_permissions' in wfas or
                          'pre_validation' in wfas):
                        adapted_level['suffix'] = 'prereviewers'
                        if 'pre_validation_keep_reviewer_permissions' in wfas:
                            adapted_level['extra_suffixes'] = ['reviewers']
                # prevalidated, disabled by default
                if adapted_level['state'] == 'prevalidated' and \
                   ('pre_validation' in wfas or
                        'pre_validation_keep_reviewer_permissions' in wfas):
                    adapted_level['enabled'] = '1'
                adapted_itemWFValidationLevels.append(adapted_level)
                cfg.setItemWFValidationLevels(adapted_itemWFValidationLevels)

    def _configureItemWFValidationLevels(self):
        """Item WF validation levels (states itemcreated, proposed, pre-validated, ...)
           are now defined in MeetingConfig.itemWFValidationLevels."""
        logger.info("Configuring 'itemWFValidationLevels' for every MeetingConfigs...")
        for cfg in self.tool.objectValues('MeetingConfig'):
            if not base_hasattr(cfg, 'historizedMeetingAttributes'):
                # historizedMeetingAttributes is removed during migration of Meeting to DX
                return self._already_migrated()
            # do apply itemWFValidation levels
            self._doConfigureItemWFValidationLevels(cfg)
            # clean stored workflowAdaptations
            cleaned_wfas = [wfa for wfa in cfg.getWorkflowAdaptations()
                            if wfa in get_vocab_values(cfg, 'WorkflowAdaptations')]
            # make sure new wfAdaptations are enabled (were default, now optional)
            cleaned_wfas += [wfa for wfa in ('pre_accepted', 'delayed', 'accepted_but_modified', )
                             if wfa in get_vocab_values(cfg, 'WorkflowAdaptations')]
            # when using "waiting_advices", and not other "waiting_advices_" wfas are enabled
            # enable the "waiting_advices_proposing_group_send_back" as it was the
            # default behaviour before and now it is configurable
            waiting_advices_wfas = [wfa for wfa in cleaned_wfas
                                    if wfa.startswith("waiting_advices")]
            if "waiting_advices" in cleaned_wfas and len(waiting_advices_wfas) == 1:
                cleaned_wfas.append("waiting_advices_proposing_group_send_back")
            # remove duplicates (in case migration is launched several times)
            cleaned_wfas = tuple(set(cleaned_wfas))
            cfg.setWorkflowAdaptations(cleaned_wfas)
            # clean some parameters that use enabled suffixes and were not validated before
            # for example even when pre_validation was not enabled, it was possible
            # to select a "prereviewers" related value, it is no more the case
            for field_name in ("itemAnnexConfidentialVisibleFor",
                               "adviceAnnexConfidentialVisibleFor",
                               "meetingAnnexConfidentialVisibleFor"):
                field = cfg.getField(field_name)
                old_value = list(field.get(cfg))
                new_value = [v for v in old_value
                             if v in field.Vocabulary(cfg)]
                if old_value != new_value:
                    self.warn(
                        logger, "Value was adapted for field \"{0}\" "
                        "of MeetingConfig \"{1}\", old_value was \"{2}\", "
                        "new_value is \"{3}\".".format(
                            field_name, cfg.getId(), old_value, new_value))
                    field.set(cfg, new_value)
            notify(ObjectEditedEvent(cfg))
        logger.info('Done.')

    def _configureVotes(self):
        """Some votes attributes on MeetingConfig already exist before changing all
           this, re-apply default values from MeetingConfigDescriptor."""
        logger.info("Configuring 'votes' for every MeetingConfigs...")
        for cfg in self.tool.objectValues('MeetingConfig'):
            if not base_hasattr(cfg, 'historizedMeetingAttributes'):
                # historizedMeetingAttributes is removed during migration of Meeting to DX
                return self._already_migrated()
            config_descr = MeetingConfigDescriptor(None, None, None)
            cfg.setUseVotes(config_descr.useVotes)
            cfg.setVotesEncoder(config_descr.votesEncoder)
            cfg.setUsedVoteValues(config_descr.usedVoteValues)
            cfg.setFirstLinkedVoteUsedVoteValues(config_descr.firstLinkedVoteUsedVoteValues)
            cfg.setNextLinkedVotesUsedVoteValues(config_descr.nextLinkedVotesUsedVoteValues)
            cfg.setVoteCondition(config_descr.voteCondition)
        # add itemVotes on every meetings, looking for meta_type='Meeting'
        # will only find AT Meeting
        brains = self.catalog(meta_type='Meeting')
        logger.info('Adding "itemVotes" to every meetings...')
        for brain in brains:
            meeting = brain.getObject()
            if hasattr(meeting, 'itemVotes'):
                continue
            meeting.itemVotes = PersistentMapping()
            # add the 'voter' key to Meeting.orderedContacts
            for value in meeting.orderedContacts.values():
                value['voter'] = False
            meeting._p_changed = True
        logger.info('Done.')

    def _updateItemSignatories(self):
        """Now that an arbitraty label may be defined when redefining item signatory,
           store it, the default value was the secondary_position_type or the position_type."""
        logger.info("Updating 'itemSignagtories' for every Meetings...")
        # looking for meta_type='Meeting' will only find AT Meeting
        brains = self.catalog(meta_type='Meeting')
        for brain in brains:
            meeting = brain.getObject()
            for item_uid, signatories in meeting.itemSignatories.items():
                for signature_number, hp_uid in signatories.items():
                    if isinstance(hp_uid, PersistentMapping):
                        continue
                    self.request.set('person_uid', hp_uid)
                    position_type = position_type_default()
                    # make sure we have a PersistentMapping
                    meeting.itemSignatories[item_uid] = PersistentMapping(
                        meeting.itemSignatories[item_uid])
                    meeting.itemSignatories[item_uid][signature_number] = \
                        PersistentMapping({'hp_uid': hp_uid, 'position_type': position_type})
        logger.info('Done.')

    def _migrateKeepAccessToItemWhenAdviceIsGiven(self):
        """Boolean field MeetingConfig.keepAccessToItemWhenAdviceIsGiven
           is now single select MeetingConfig.keepAccessToItemWhenAdvice."""
        logger.info(
            "Migrationg field 'keepAccessToItemWhenAdviceIsGiven' for "
            "every MeetingConfigs and organizations...")
        # MeetingConfigs
        for cfg in self.tool.objectValues('MeetingConfig'):
            if hasattr(cfg, 'keepAccessToItemWhenAdviceIsGiven'):
                old_value = cfg.keepAccessToItemWhenAdviceIsGiven
                if old_value is True:
                    cfg.setKeepAccessToItemWhenAdvice('is_given')
                delattr(cfg, 'keepAccessToItemWhenAdviceIsGiven')
        # organizations
        orgs = get_organizations(only_selected=False)
        for org in orgs:
            # org.keep_access_to_item_when_advice_is_given could not exist
            # for imported org never saved
            old_value = getattr(org, 'keep_access_to_item_when_advice_is_given', None)
            if old_value == '':
                org.keep_access_to_item_when_advice = 'use_meetingconfig_value'
            elif old_value == '1':
                org.keep_access_to_item_when_advice = 'is_given'
            elif old_value == '0':
                org.keep_access_to_item_when_advice = 'default'
        logger.info('Done.')

    def _removeMeetingItemsReferenceField(self):
        '''ReferenceField Meeting.items was removed and is now managed manually.'''
        logger.info("Removing Meeting.items reference field...")
        # looking for meta_type='Meeting' will only find AT Meeting
        for brain in self.catalog(meta_type='Meeting'):
            meeting = brain.getObject()
            # get references from at_references so order is kept
            reference_uids = [ref.targetUID for ref in meeting.at_references.objectValues()
                              if ref.relationship == 'MeetingItems']
            if reference_uids:
                meeting.deleteReferences('MeetingItems')
                brains = self.catalog(UID=reference_uids)
                for brain in brains:
                    item = brain.getObject()
                    item._update_meeting_link(meeting)
        logger.info('Done.')

    def _updateItemPreferredMeetingLink(self):
        """Update MeetingItem.preferred_meeting_path for every items."""
        logger.info("Updating MeetingItem.preferred_meeting_path for every items...")
        pghandler = ZLogHandler(steps=1000)
        brains = self.catalog(meta_type='MeetingItem')
        pghandler.init('Updating MeetingItem.preferred_meeting_path for every items...',
                       len(brains))
        i = 0
        for brain in brains:
            i += 1
            pghandler.report(i)
            item = brain.getObject()
            item._update_preferred_meeting(item.getPreferredMeeting())
        pghandler.finish()
        logger.info('Done.')

    def _fixRichTextValueMimeType(self,
                                  object_provides=IMeetingAdvice.__identifier__,
                                  field_names=['advice_comment', 'advice_observations']):
        """Make sure RichTextValue stored on DX content (advices) have
           a correct mimeType and outputMimeType."""
        brains = self.catalog(object_provides=object_provides)
        logger.info('Fixing mimeType/outputMimeType for every %s...' % object_provides)
        pghandler = ZLogHandler(steps=1000)
        pghandler.init('Updating RichText fields for elements...', len(brains))
        i = 0
        for brain in brains:
            i += 1
            pghandler.report(i)
            try:
                obj = brain.getObject()
            except AttributeError:
                continue
            for field_name in field_names:
                field_value = getattr(obj, field_name)
                if field_value:
                    setattr(obj, field_name, richtextval(field_value.raw))
        pghandler.finish()
        logger.info('Done.')

    def _updateSearchedFolderBatchActionsMarkerInterface(self):
        """Update every MeetingConfig batch actions marker applied
           to sub folders, now there is a different marker for
           dashboards displaying items or meetings."""
        logger.info("Updating every meeting folders batch actions marker interfaces...")
        for cfg in self.tool.objectValues('MeetingConfig'):
            folders = cfg._get_all_meeting_folders()
            for folder in folders:
                for sub_folder in folder.objectValues('ATFolder'):
                    if not sub_folder.getId().startswith('searches_') or \
                       not IBatchActionsMarker.providedBy(sub_folder):
                        continue
                    noLongerProvides(sub_folder, IBatchActionsMarker)
                    if sub_folder.getId() == "searches_items":
                        # item related searches
                        alsoProvides(sub_folder, IMeetingItemDashboardBatchActionsMarker)
                    else:
                        # meeting related searches
                        alsoProvides(sub_folder, IMeetingDashboardBatchActionsMarker)
                    sub_folder.reindexObject(idxs=['object_provides'])
        logger.info('Done.')

    def _fixFacetedFilters(self):
        '''Filter modified(c13)/created(c14) is now created(c13)/modified(c14).
           Filter hasAnnexesToPrint(c20) and hasAnnexesToSign(c25) are merged to (c20)
           and will use the annexes_index.'''
        logger.info('Fixing created/modified and hasAnnexesToPrint/hasAnnexesToSign filters...')
        for cfg in self.tool.objectValues('MeetingConfig'):
            # make sure only done one time or relaunching migration would
            # reapply and switch back to wrong configuration
            if not base_hasattr(cfg, 'historizedMeetingAttributes'):
                # historizedMeetingAttributes is removed during migration of Meeting to DX
                return self._already_migrated()
            for field_name in ('dashboardItemsListingsFilters',
                               'dashboardMeetingAvailableItemsFilters',
                               'dashboardMeetingLinkedItemsFilters'):
                field = cfg.getField(field_name)
                keys = list(field.get(cfg))
                # created/modified
                if 'c13' in keys and 'c14' in keys:
                    # nothing to do as both were already selected
                    pass
                elif 'c13' in keys:
                    keys = replace_in_list(keys, 'c13', 'c14')
                elif 'c14' in keys:
                    keys = replace_in_list(keys, 'c14', 'c13')
                # annexes_index 'c25' was removed and merged in 'c20'
                if 'c25' in keys:
                    keys.remove('c25')
                    if 'c20' not in keys:
                        keys.append('c20')
                field.set(cfg, sorted(keys))

        logger.info('Fixing orgs-searches review_sate filters...')
        orgs_searches_folder = self.portal.contacts.get('orgs-searches')
        orgs_searches_folder_criteria = ICriteria(orgs_searches_folder)
        active_org_criterion = orgs_searches_folder_criteria.get('c6')
        active_org_criterion.hidden = True
        cleanMemoize(self.portal, prefixes=['plonegroupl-utils-get_organizations-'])
        logger.info('Done.')

    def _migrateItemPredecessorReference(self):
        '''MeetingItem.predecessor ReferenceField is managed manually now.'''
        logger.info("Migrating MeetingItem.predecessor ReferenceField...")
        pghandler = ZLogHandler(steps=1000)
        brains = self.portal.reference_catalog(relationship='ItemPredecessor')
        pghandler.init('Migrating MeetingItem.predecessor reference field...', len(brains))
        pghandler.info('Migrating MeetingItem.predecessor reference field...')
        i = 0
        for brain in brains:
            i += 1
            pghandler.report(i)
            relation = brain.getObject()
            if not relation:
                self.warn(
                    logger,
                    'In _migrateItemPredecessorReference, no relation found for "{0}"'.format(
                        brain.UID))
                continue
            item = relation.getSourceObject()
            predecessor = relation.getTargetObject()
            item._update_predecessor(predecessor)
        # deleteReferences in a second phase
        for brain in brains:
            relation = brain.getObject()
            if relation:
                item = relation.getSourceObject()
                item.deleteReferences('ItemPredecessor')
        pghandler.finish()
        logger.info('Done.')

    def _updateConfigForAdviceAskedAgainNoMoreOptional(self):
        '''Advice type "asked_again" is no more optional, fix MeetingConfigs.'''
        logger.info('Updating every MeetingConfigs advice_type related parameters....')
        for cfg in self.tool.objectValues('MeetingConfig'):
            usedAdviceTypes = list(cfg.getUsedAdviceTypes())
            if "asked_again" in usedAdviceTypes:
                usedAdviceTypes.remove("asked_again")
                cfg.setUsedAdviceTypes(usedAdviceTypes)
            defaultAdviceType = cfg.getDefaultAdviceType()
            if defaultAdviceType == "asked_again":
                defaultAdviceType = "positive"
                cfg.setDefaultAdviceType(defaultAdviceType)
        logger.info('Done.')

    def _updatePortletTodoTitleLength(self):
        """Set title_length in portlet_todo to "100" instead "60"."""
        logger.info('Updating portlet_todo.title_length to "100"...')
        portal_ann = IAnnotations(self.portal)
        left_col = portal_ann["plone.portlets.contextassignments"]["plone.leftcolumn"]
        if "portlet_todo" not in left_col:
            self.warn(logger, 'Could not find "portlet_todo" at root of Plone Site!')
        portlet_todo = left_col["portlet_todo"]
        portlet_todo.title_length = 100
        logger.info('Done.')

    def _cleanUnusedPersonsAndHeldPositions(self):
        """Persons and HeldPositions migrated from old MeetingUsers and not
           used will be removed."""

        class DummyEvent(object):
            def __init__(self, obj):
                self.object = obj

        logger.info('Cleaning up unused "Persons" and "HeldPositions"...')
        # as we could remove objects, use an intermediate list of paths
        # because we can not iterate a list in which an element was deleted
        # (list of brains of objects)
        paths = [brain.getPath() for brain in
                 self.catalog.unrestrictedSearchResults(portal_type="held_position")]
        for path in paths:
            hp = self.portal.restrictedTraverse(path)
            if hp.getId().endswith('_hp1'):
                dummy_event = DummyEvent(hp)
                try:
                    onHeldPositionWillBeRemoved(hp, dummy_event)
                    person = hp.get_person()
                    # remove the person if it only contains this hp
                    if len(person.objectIds()) == 1:
                        api.content.delete(person)
                    else:
                        api.content.delete(hp)
                    self.warn(
                        logger,
                        'Directory person or held_position at "{0}" was removed!'.format(
                            "/".join(person.getPhysicalPath())))
                except BeforeDeleteException:
                    continue
        logger.info('Done.')

    def _updateMeetingsNumberOfItems(self):
        """Meeting number of items is now stored in Meeting._number_of_items."""
        logger.info('Updating "_number_of_items" for every meetings...')
        brains = self.catalog(object_provides=IMeeting.__identifier__)
        for brain in brains:
            meeting = brain.getObject()
            meeting._number_of_items = len(meeting.get_raw_items())
            meeting._p_changed = True
        logger.info('Done.')

    def _updateItemGroupsInCharge(self):
        '''When using MetingConfig.includeGroupsInChargeDefinedOnProposingGroup or
           MeetingConfig.includeGroupsInChargeDefinedOnCategory, for history reason,
           we store the resulting groupsInCharge if MeetingItem.groupsInCharge is empty.'''
        logger.info("Updating MeetingItem.groupsInCharge...")
        for cfg in self.tool.objectValues('MeetingConfig'):
            if cfg.getIncludeGroupsInChargeDefinedOnProposingGroup() or \
               cfg.getIncludeGroupsInChargeDefinedOnCategory():
                i = 0
                brains = self.catalog(portal_type=cfg.getItemTypeName(configType='all'))
                msg = 'Updating items for MeetingConfig "{0}"...'.format(cfg.Title())
                pghandler = ZLogHandler(steps=1000)
                pghandler.init(msg, len(brains))
                pghandler.info(msg)
                for brain in brains:
                    i += 1
                    pghandler.report(i)
                    item = brain.getObject()
                    item.update_groups_in_charge()
                pghandler.finish()
        logger.info('Done.')

    def _fixPODTemplatesInstructions(self):
        '''Make some replace in POD templates to fit changes in code...'''
        # for every POD templates
        replacements = {'listTypes=': 'list_types=',
                        '.getDate()': '.date',
                        '.getStartDate()': '.start_date',
                        '.getMidDate()': '.mid_date',
                        '.getEndDate()': '.end_date',
                        '.getApprovalDate()': '.approval_date',
                        '.getConvocationDate()': '.convocation_date',
                        '.getPreMeetingDate()': '.pre_meeting_date',
                        '.getPlace()': '.get_place()',
                        '.getExtraordinarySession()': ".extraordinary_session",
                        '.getFirstItemNumber()': ".first_item_number",
                        '.getAttendees(theObjects': '.get_attendees(the_objects',
                        '.getAttendees(': '.get_attendees(',
                        '.getAbsents(theObjects': '.get_absents(the_objects',
                        '.getAbsents(': '.get_absents(',
                        '.getExcused(theObjects': '.get_excused(the_objects',
                        '.getExcused(': '.get_excused(',
                        '.getItemAbsents(theObjects': '.get_item_absents(the_objects',
                        '.getItemAbsents(': '.get_item_absents(',
                        '.getItemExcused(theObjects': '.get_item_excused(the_objects',
                        '.getItemExcused(': '.get_item_excused(',
                        '.getItemNonAttendees(theObjects': '.get_item_non_attendees(the_objects',
                        '.getItemNonAttendees(': '.get_item_non_attendees(',
                        '.getAssembly(': '.get_assembly(',
                        '.getAssemblyAbsents(': '.get_assembly_absents(',
                        '.getAssemblyExcused(': '.get_assembly_excused(',
                        '.getAssemblyGuests(': '.get_assembly_guests(',
                        '.getAssemblyStaves(': '.get_assembly_staves(',
                        '.getAssemblyProxies(': '.get_assembly_proxies(',
                        '.getSignatories(theObjects=': '.get_signatories(the_objects=',
                        '.getSignatories(': '.get_signatories(',
                        '.getMeetingNumber()': ".meeting_number",
                        '.getItems(': '.get_items(',
                        '.getItemsInOrder(': '.get_items(ordered=True, ',
                        'self.adapted().getPrintableItems(itemUids)':
                            'self.get_items(uids=uids, ordered=True)',
                        '.getItemSignatories(theObjects=': '.get_item_signatories(the_objects=',
                        '.getItemSignatories(': '.get_item_signatories(',
                        '.getNextMeeting(cfgId=': '.get_next_meeting(cfg_id=',
                        '.getNextMeeting(': '.get_next_meeting(',
                        '.getMeetingNumber()': ".meeting_number",
                        # get_next_meeting parameter
                        'dateGap=': 'date_gap=',
                        '.numberOfItems(': '.number_of_items(',
                        '.queryState(': '.query_state(',
                        'zamqp_utils.scan_id_barcode(self,': 'view.print_scan_id_barcode(',
                        '.printAdvicesInfos(': '.print_advices_infos(',
                        '.printAllAnnexes(': '.print_all_annexes(',
                        '.printAssembly(': '.print_assembly(',
                        '.printFinanceAdvice(': '.print_finance_advice(',
                        '.printFormatedAdvice(': '.print_formated_advice(',
                        '.printFullname(': '.print_fullname(',
                        '.printHistory(': '.print_history(',
                        '.printMeetingDate(': '.print_meeting_date(',
                        '=imageOrientation': '=image_orientation',
                        'self.getStrikedAssembly(groupByDuty=True': 'view.print_assembly(',
                        'self.getStrikedItemAssembly(groupByDuty=True': 'view.print_assembly(',
                        'self.getStrikedAssembly(': 'view.print_assembly(',
                        'self.getStrikedItemAssembly(': 'view.print_assembly(',
                        '.isDecided(': '.is_decided(',
                        # formatMeetingDate to format_date
                        'withHour=': "with_hour=",
                        'withWeekDayName=': "with_week_day_name=",
                        # meeting date month name without accents
                        ".replace('é','e').replace('û','u').upper()":
                            ".replace(u'é',u'e').replace(u'û',u'u').upper()",
                        # uid_catalog can no more be used to get DX Meeting
                        ".uid_catalog(": ".portal_catalog(",
                        # catalog is now available in default context
                        "self.portal_catalog(": "catalog(",
                        # get_assembly, striked=True by default
                        # also used in some dashboard POD templates
                        '.displayStrikedAssembly()': '.get_assembly()',
                        # called on a MeetingCategory
                        '.getCategoryId()': '.category_id',
                        }
        # specific for Meeting POD Templates
        meeting_replacements = {
            'self.getAuthorityNotice()': "view.print_value('authority_notice')",
            'self.getCommitteeObservations()': "view.print_value('committees_observations')",
            'self.getInAndOutMoves()': "view.print_value('in_and_out_moves')",
            'self.getNotes()': "view.print_value('notes')",
            'self.getObservations()': "view.print_value('observations')",
            'self.getPlace()': "view.print_value('place')",
            'self.getPreObservations()': "view.print_value('pre_observations')",
            'self.getPublicMeetingObservations()': "view.print_value('public_meeting_observations')",
            'self.getSecretMeetingObservations()': "view.print_value('secret_meeting_observations')",
            'self.getSignatures()': "self.get_signatures()",
            'self.Title()': "tool.format_date(self.date)",
            # formatMeetingDate to format_date
            'tool.formatMeetingDate(self': "tool.format_date(self.date",
            # display_date
            ".display_date('startDate'": ".display_date('start_date'",
            ".display_date('midDate'": ".display_date('mid_date'",
            ".display_date('endDate'": ".display_date('end_date'",
            ".display_date('approvalDate'": ".display_date('approval_date'",
            ".display_date('convocationDate'": ".display_date('convocation_date'",
            ".display_date('preMeetingDate'": ".display_date('pre_meeting_date'",
            ".display_date('approvalDate'": ".display_date('approval_date'",
        }
        # specific for MeetingItem POD Templates
        item_replacements = {
            'meeting.Title()': "tool.format_date(meeting.date)",
            'self.getMeeting().Title()': "tool.format_date(self.getMeeting().date)",
            # formatMeetingDate to format_date
            'tool.formatMeetingDate(meeting': "tool.format_date(meeting.date",
            'tool.formatMeetingDate(self.getMeeting()': "tool.format_date(self.getMeeting().date",
            # getItemAssembly, striked=True by default
            '.displayStrikedItemAssembly()': '.getItemAssembly()',
            # used in Avis DF
            'self.displayValue(self.listProposingGroups(), self.getProposingGroup())':
                "view.display('proposingGroup')",
            # meeting is now available in default generation context
            'self.getMeeting()': 'meeting',
        }
        self.updatePODTemplatesCode(replacements, meeting_replacements, item_replacements)

    def _fixItemAddAdvicePermission(self):
        """Changed role that is able to add advice from 'Contributor' to 'MeetingAdviser'.
           Actually we just remove the 'Contributor' role from 'PloneMeeting: Add advice' permission
           as the update local roles we set it back again correctly."""
        logger.info("Removing role 'Contributor' for add advice permission for every items...")
        for brain in self.catalog(meta_type='MeetingItem'):
            item = brain.getObject()
            item._removePermissionToRole(
                permission=AddAdvice,
                role_to_remove='Contributor',
                obj=item)
        logger.info('Done.')

    def _updateCron4Plone(self):
        """The maintenance task view name changed to @@pm-night-tasks."""
        logger.info("Updating cron4plone configuration...")
        cron_configlet = queryUtility(ICronConfiguration, 'cron4plone_config')
        cron_configlet.cronjobs = [u'45 1 * * portal/@@pm-night-tasks']
        logger.info('Done.')

    def _initMeetingConfigItemInternalNotesEditableBy(self):
        """By default, make proposingGroup editors able to use MeetingItem.internalNotes."""
        logger.info("Updating every MeetingConfig.ItemInternalNotesEditableBy...")
        for cfg in self.tool.objectValues('MeetingConfig'):
            # update if used and if not already migrated
            if "internalNotes" in cfg.getUsedItemAttributes() and \
                    not cfg.getItemInternalNotesEditableBy():
                suffixes = cfg.getItemWFValidationLevels(data='suffix', only_enabled=True)
                values = ['{0}{1}'.format(PROPOSINGGROUPPREFIX, suffix)
                          for suffix in suffixes]
                cfg.setItemInternalNotesEditableBy(values)
        logger.info('Done.')

    def _correctAccessToPODTemplates(self):
        """Correct weird Unauthorized when accessing POD template file."""
        logger.info('Correcting access to POD Templates file...')
        brains = self.catalog.unrestrictedSearchResults(
            object_provides=(
                'collective.documentgenerator.content.pod_template.IPODTemplate',
                'collective.documentgenerator.content.style_template.IStyleTemplate'))
        for brain in brains:
            pod_template = brain._unrestrictedGetObject()
            logger.info('Updating access for POD template at {0}'.format(
                '/'.join(pod_template.getPhysicalPath())))
            pod_template.reindexObject()
            pod_template.reindexObjectSecurity()
        logger.info('Done.')

    def _fixMeetingConfigRelatedPloneGroupsLocalRoles(self):
        '''When a MeetingConfig is created, some local_roles are set,
           especially on contacts, make sure it is correct, it could happen
           with old code that there were missing local_roles.'''
        logger.info("Updating MeetingConfig related Plone groups access...")
        for cfg in self.tool.objectValues('MeetingConfig'):
            logger.info('Migrating config {0}...'.format(cfg.getId()))
            # give local_roles to relevant groups on contacts
            cfg._createOrUpdateAllPloneGroups(force_update_access=True)
        logger.info('Done.')

    def run(self, extra_omitted=[]):
        logger.info('Migrating to PloneMeeting 4200...')

        if self.is_in_part('a'):  # main step, everything but update local roles and refresh catalog
            self._fixPODTemplatesInstructions()
            self._fixFacetedFilters()

            # apply correct batch actions marker on searches_* folders
            self._updateSearchedFolderBatchActionsMarkerInterface()

            # update cron4plone
            self._updateCron4Plone()

            # update preferred meeting path on items
            self._updateItemPreferredMeetingLink()
            self._migrateItemPredecessorReference()
            self._updateConfigForAdviceAskedAgainNoMoreOptional()
            self._updateItemGroupsInCharge()
            self._correctAccessToPODTemplates()

            # remove useless catalog indexes and columns, were renamed to snake case
            self.removeUnusedIndexes(
                indexes=['getItemIsSigned',
                         'sendToAuthority',
                         'toDiscuss',
                         'getDate',
                         'linkedMeetingUID',
                         'linkedMeetingDate',
                         'hasAnnexesToPrint',
                         'hasAnnexesToSign',
                         'item_boolean_indexes'])
            self.removeUnusedColumns(
                columns=['toDiscuss',
                         'getDate',
                         'getGroupInCharge',
                         'getItemNumber',
                         'linkedMeetingUID',
                         'linkedMeetingDate'])

            # need to update schema policies manually before reinstall
            # because GS will look for old schema_policy during update and it
            # does not exist anymore (category_zamqp_schema_policy)
            migrate_to_4209 = Migrate_To_4209(self.portal)
            migrate_to_4209._updateContentCategoryPortalTypes()

            # reinstall workflows before updating workflowAdaptations
            self.runProfileSteps('Products.PloneMeeting', steps=['workflow'], profile='default')
            # make sure new portal_type Meeting is installed
            self.removeUnusedPortalTypes(portal_types=['Meeting'])
            self.ps.runImportStepFromProfile('profile-Products.PloneMeeting:default', 'typeinfo')
            # configure wfAdaptations before reinstall
            self._configureItemWFValidationLevels()

            # remove broken annexes after item WF update
            # because we need item.query_state and it will only work if item WF ready
            self._removeBrokenAnnexes()

            # init MeetingConfig.itemInternalNotesEditableBy after _configureItemWFValidationLevels
            self._initMeetingConfigItemInternalNotesEditableBy()

            # need to reindex new indexes before migrating Meeting to DX
            addOrUpdateIndexes(self.portal, indexInfos)
            addOrUpdateColumns(self.portal, columnInfos, update_metadata=False)

            # update various TAL expressions
            self.updateTALConditions("queryState", "query_state")
            self.updateTALConditions("updateLocalRoles", "update_local_roles")
            self.updateTALConditions("getDate()", "date")
            self.updateTALConditions("getStartDate()", "start_date")
            self.updateTALConditions("getEndDate()", "end_date")
            self.updateTALConditions(".getMeetingNumber()", ".meeting_number")
            self.updateTALConditions("member.getGroups()", "tool.get_plone_groups_for_user()")
            self.updateTALConditions("power_observer_type='restrictedpowerobservers')",
                                     "power_observer_types=['restrictedpowerobservers'])")
            self.updateTALConditions("power_observer_type='powerobservers')",
                                     "power_observer_types=['powerobservers'])")
            self.updateTALConditions("isManager(context)",
                                     "isManager(cfg")
            self.updateTALConditions("isManager(here)",
                                     "isManager(cfg)")
            self.updateTALConditions("isManager(obj)",
                                     "isManager(cfg)")
            self.updateTALConditions("isManager(item)",
                                     "isManager(cfg)")

            self.updateTALConditions("isManager(context, realManagers=True)",
                                     "isManager(realManagers=True)")
            self.updateTALConditions("isManager(here, realManagers=True)",
                                     "isManager(realManagers=True)")
            self.updateTALConditions("isManager(obj, realManagers=True)",
                                     "isManager(realManagers=True)")
            self.updateTALConditions("isManager(tool, realManagers=True)",
                                     "isManager(realManagers=True)")
            self.updateTALConditions("isManager(item, realManagers=True)",
                                     "isManager(realManagers=True)")

            self.updateTALConditions("isManager(context, True)",
                                     "isManager(realManagers=True)")
            self.updateTALConditions("isManager(here, True)",
                                     "isManager(realManagers=True)")
            self.updateTALConditions("isManager(obj, True)",
                                     "isManager(realManagers=True)")
            self.updateTALConditions("isManager(tool, True)",
                                     "isManager(realManagers=True)")
            self.updateTALConditions("isManager(item, True)",
                                     "isManager(realManagers=True)")

            self.updateTALConditions("isManager(context,realManagers=True)",
                                     "isManager(realManagers=True)")
            self.updateTALConditions("isManager(here,realManagers=True)",
                                     "isManager(realManagers=True)")
            self.updateTALConditions("isManager(obj,realManagers=True)",
                                     "isManager(realManagers=True)")
            self.updateTALConditions("isManager(tool,realManagers=True)",
                                     "isManager(realManagers=True)")
            self.updateTALConditions("isManager(context,True)",
                                     "isManager(realManagers=True)")
            self.updateTALConditions("isManager(here,True)",
                                     "isManager(realManagers=True)")
            self.updateTALConditions("isManager(obj,True)",
                                     "isManager(realManagers=True)")
            self.updateTALConditions("isManager(tool,True)",
                                     "isManager(realManagers=True)")

            self.updateTALConditions(
                "'pre_validation' in cfg.getWorkflowAdaptations()",
                "'prevalidated' in cfg.getItemWFValidationLevels(data='state', only_enabled=True)")
            self.updateTALConditions(".showHolidaysWarning(context)", ".showHolidaysWarning(cfg)")

            # replacements MeetingConfig item columns
            self.updateColumns(
                to_replace={'getPreferredMeetingDate': 'preferred_meeting_date',
                            'linkedMeetingDate': 'meeting_date'})

            self._migrateKeepAccessToItemWhenAdviceIsGiven()

            # MEETING TO DX
            self._hook_before_meeting_to_dx()
            self._migrateMeetingToDX()
            self._hook_after_meeting_to_dx()

            # update RichTextValue stored on DX types (advices)
            self._fixRichTextValueMimeType()

            self.upgradeAll(omit=['Products.PloneMeeting:default',
                                  self.profile_name.replace('profile-', '')] + extra_omitted)

            # reinstall so versions are correctly shown in portal_quickinstaller
            self.reinstall(profiles=['profile-Products.PloneMeeting:default', ],
                           ignore_dependencies=False,
                           dependency_strategy=DEPENDENCY_STRATEGY_NEW)
            if self.profile_name != 'profile-Products.PloneMeeting:default':
                self.reinstall(profiles=[self.profile_name, ],
                               ignore_dependencies=False,
                               dependency_strategy=DEPENDENCY_STRATEGY_NEW)

            # configure new WFs
            self.cleanMeetingConfigs(field_names=['itemDecidedStates', 'itemPositiveDecidedStates'])

            # init otherMeetingConfigsClonableToFieldXXX and XXXSuite/XXXEnd new fields
            self.initNewHTMLFields(query={'meta_type': ('MeetingItem')})

            # reimport every advanced widgets (so except c0/c1/c2/c3)
            self.updateFacetedFilters(xml_filename='upgrade_step_4200_add_item_widgets.xml')
            self.updateFacetedFilters(
                xml_filename='upgrade_step_4200_update_meeting_widgets.xml',
                related_to="meetings")
            self.updateFacetedFilters(
                xml_filename='upgrade_step_4200_update_meeting_widgets.xml',
                related_to="decisions")
            self.updateFacetedFilters(reorder=False, to_delete=['c25'])

            # set title_length=100 for portlet_todo
            self._updatePortletTodoTitleLength()

            # update holidays
            self.updateHolidays()

            # add new collections, the "searchmyitemstoadvice" for example
            self.addNewSearches()

            # adviser role able to add advice is now MeetingAdviser
            self._fixItemAddAdvicePermission()

            # add the Optimizate columns CKeditor style
            self.addCKEditorStyle("table_optimization", "table", "style", "table-layout:auto;")

        if self.is_in_part('b'):  # update_all_local_roles
            # update workflow mappings before local_roles because permissions to add
            # elements like "Image" will use it
            self.refreshDatabase(catalogs=False, workflows=True)
            self.warnings += self.tool.update_all_local_roles(redirect=False)

        if self.is_in_part('c'):  # refresh catalog and workflow mappings
            self.refreshDatabase(catalogsToUpdate=[])

            # store meeting number of items
            self._updateMeetingsNumberOfItems()

            # remove unused persons from contacts directory
            self._cleanUnusedPersonsAndHeldPositions()

            # make sure MeetingConfig related Plone groups local_roles are correct
            self._fixMeetingConfigRelatedPloneGroupsLocalRoles()


def migrate(context):
    '''This migration function will:

       1) Fix faceted filters;
       2) Update applied batch actions marker interface on every member folders;
       3) Update preferredMeeting behavior;
       4) Remove and migrate item predecessor;
       5) Update MeetingConfigs as advice type "asked_again" is no more optional;
       6) Remove unused indexes and metadata;
       7) Remove Meeting.items reference field;
       8) Configure votes;
       9) Update Meeting.itemSignatories to manage stored position_type;
       10) Fix DX RichText mimetype;
       11) Configure field MeetingConfig.itemWFValidationLevels depending on old wfAdaptations;
       12) Migrate MeetingConfig.keepAccessToItemWhenAdviceIsGiven to
          MeetingConfig.keepAccessToItemWhenAdvice;
       13) Init otherMeetingConfigsClonableToFieldXXX new fields;
       14) Update faceted filters;
       15) Update holidays;
       16) Refresh items local roles and recatalog.
    '''
    migrator = Migrate_To_4200(context)
    migrator.run()
    migrator.finish()



# -*- coding: utf-8 -*-
#
# GNU General Public License (GPL)
#

from AccessControl import Unauthorized
from collective.contact.plonegroup.utils import get_plone_group_id
from imio.helpers.cache import get_current_user_id
from imio.helpers.cache import get_plone_groups_for_user
from imio.helpers.content import get_vocab
from plone import api
from plone.directives import form
from plone.z3cform.layout import wrap_form
from Products.CMFPlone import PloneMessageFactory as PMF
from Products.PloneMeeting.config import DUPLICATE_AND_KEEP_LINK_EVENT_ACTION
from Products.PloneMeeting.config import DUPLICATE_EVENT_ACTION
from Products.PloneMeeting.config import PMMessageFactory as _
from Products.PloneMeeting.interfaces import IRedirect
from Products.PloneMeeting.widgets.pm_checkbox import PMCheckBoxFieldWidget
from z3c.form import button
from z3c.form import field
from z3c.form import form as z3c_form
from z3c.form.browser.radio import RadioFieldWidget
from zope import schema
from zope.component import queryUtility
from zope.i18n import translate
from zope.interface import provider
from zope.schema._bootstrapinterfaces import IContextAwareDefaultFactory
from zope.security.interfaces import IPermission


@provider(IContextAwareDefaultFactory)
def annex_ids_default(context):
    """Select every annexes by default."""
    vocab = get_vocab(
        context,
        u"Products.PloneMeeting.vocabularies.item_duplication_contained_annexes_vocabulary")
    return vocab.by_token.keys()


class IDuplicateItem(form.Schema):
    """ """

    keep_link = schema.Bool(
        title=_(u'Keep link?'),
        description=_(""),
        required=False,
        default=False,
    )

    annex_ids = schema.List(
        title=_(u"Annexes to keep"),
        description=_(u""),
        required=False,
        defaultFactory=annex_ids_default,
        value_type=schema.Choice(
            vocabulary=u"Products.PloneMeeting.vocabularies.item_duplication_contained_annexes_vocabulary"),
    )

    annex_decision_ids = schema.List(
        title=_(u"Decision annexes to keep"),
        description=_(u""),
        required=False,
        value_type=schema.Choice(
            vocabulary=u"Products.PloneMeeting.vocabularies.item_duplication_contained_decision_annexes_vocabulary"),
    )


class DuplicateItemForm(z3c_form.Form):
    """ """
    fields = field.Fields(IDuplicateItem)
    fields["keep_link"].widgetFactory = RadioFieldWidget
    fields["annex_ids"].widgetFactory = PMCheckBoxFieldWidget
    fields["annex_decision_ids"].widgetFactory = PMCheckBoxFieldWidget

    ignoreContext = True  # don't use context to get widget data

    label = PMF(u"Duplicate")
    description = _('Disabled (greyed) annexes will not be kept on the new duplicated item.')
    _finished = False

    def __init__(self, context, request):
        self.context = context
        self.request = request

    @button.buttonAndHandler(_('Apply'), name='apply_duplicate_item')
    def handleApply(self, action):
        self._check_auth()
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        self._doApply(data)

    def _check_data(self, data):
        """Make sure annex_ids/annex_decision_ids are correct.
           As some values are disabled in the UI, a user could try
           to surround this, raise Unauthorized in this case."""
        annex_terms = self.widgets['annex_ids'].terms
        annex_term_ids = [term.token for term in annex_terms
                          if not term.disabled]
        for annex_id in data['annex_ids']:
            if annex_id not in annex_term_ids:
                raise Unauthorized
        decision_annex_terms = self.widgets['annex_decision_ids'].terms
        decision_annex_term_ids = [term.token for term in decision_annex_terms
                                   if not term.disabled]
        for decision_annex_id in data['annex_decision_ids']:
            if decision_annex_id not in decision_annex_term_ids:
                raise Unauthorized

    def _doApply(self, data):
        """ """
        user_id = get_current_user_id()
        cloneEventAction = DUPLICATE_EVENT_ACTION
        setCurrentAsPredecessor = False
        manualLinkToPredecessor = False
        if data['keep_link']:
            cloneEventAction = DUPLICATE_AND_KEEP_LINK_EVENT_ACTION
            setCurrentAsPredecessor = True
            manualLinkToPredecessor = True
        # make sure data is correct
        self._check_data(data)

        # as passing empty keptAnnexIds/keptDecisionAnnexIds ignores it
        # if we unselect every annexes, we force copyAnnexes/copyDecisionAnnexes to False
        copyAnnexes = data['annex_ids'] and True or False
        copyDecisionAnnexes = data['annex_decision_ids'] and True or False
        # keep proposingGroup if current user creator for it
        keepProposingGroup = False
        proposingGroup = self.context.getProposingGroup()
        if get_plone_group_id(proposingGroup, 'creators') in get_plone_groups_for_user():
            keepProposingGroup = True
        newItem = self.context.clone(
            copyAnnexes=copyAnnexes,
            copyDecisionAnnexes=copyDecisionAnnexes,
            newOwnerId=user_id,
            cloneEventAction=cloneEventAction,
            keepProposingGroup=keepProposingGroup,
            setCurrentAsPredecessor=setCurrentAsPredecessor,
            manualLinkToPredecessor=manualLinkToPredecessor,
            keptAnnexIds=data['annex_ids'],
            keptDecisionAnnexIds=data['annex_decision_ids'])
        self.new_item_url = newItem.absolute_url()
        api.portal.show_message(
            translate('item_duplicated', domain='PloneMeeting', context=self.request),
            request=self.request)
        self._finished = True
        return newItem

    @button.buttonAndHandler(_('Cancel'), name='cancel')
    def handleCancel(self, action):
        self._finished = True

    def update(self):
        """ """
        self._check_auth()
        super(DuplicateItemForm, self).update()
        # after calling parent's update, self.actions are available
        self.actions.get('cancel').addClass('standalone')

    def updateWidgets(self):
        # XXX manipulate self.fields BEFORE doing form.Form.updateWidgets
        z3c_form.Form.updateWidgets(self)
        # disable values if current user not able to add annex or annexDecision
        if not self._user_able_to_add():
            for term in self.widgets['annex_ids'].terms:
                term.disabled = True
                term.title = term.title + u' [Not allowed by application permissions]'
        elif not self._user_able_to_add(annex_portal_type='annexDecision'):
            for term in self.widgets['annex_decision_ids'].terms:
                term.disabled = True
                term.title = term.title + u' [Not allowed by application permissions]'

    def _user_able_to_add(self, annex_portal_type='annex'):
        """Is current user able to add given p_annex_portal_type
           on newly created item?"""
        # does user have Add annex when item created?
        typesTool = api.portal.get_tool('portal_types')
        wfTool = api.portal.get_tool('portal_workflow')
        item_wf = wfTool.getWorkflowsFor(self.context)[0]
        initial_state = item_wf.states.get(item_wf.initial_state)
        add_permission = typesTool.get(annex_portal_type).add_permission
        # add_permission is a z3 like permission (PloneMeeting.AddAnnex),
        # need the permission title (u'PloneMeeting: Add annex')
        add_permission = queryUtility(IPermission, add_permission).title
        permission_roles = initial_state.permission_roles[add_permission]
        res = False
        # Editor is the role that may add annex, Contributor is used to add annex decision
        if 'Editor' in permission_roles or \
           'Contributor' in permission_roles:
            res = True
        return res

    def _check_auth(self):
        """Raise Unauthorized if current user may not duplicate the item."""
        if not self.context.showDuplicateItemAction():
            raise Unauthorized

    def render(self):
        if self._finished:
            IRedirect(self.request).redirect(
                getattr(self, 'new_item_url', self.context.absolute_url()))
            return ""
        return super(DuplicateItemForm, self).render()


DuplicateItemFormWrapper = wrap_form(DuplicateItemForm)



# -*- coding: utf-8 -*-
#
# File: testing.py
#
# Copyright (c) 2015 by Imio.be
#
# GNU General Public License (GPL)
#

import monkey  # flake8: noqa
from plone import api
from plone.app.robotframework.testing import REMOTE_LIBRARY_BUNDLE_FIXTURE
from plone.app.testing import FunctionalTesting
from plone.app.testing import IntegrationTesting
from plone.app.testing import PloneWithPackageLayer
from plone.app.testing.bbb import _createMemberarea
from plone.testing import z2
from plone.testing import zca
from Products.CMFPlone.utils import base_hasattr
from zope.globalrequest.local import setLocal

import Products.PloneMeeting


class PMLayer(PloneWithPackageLayer):

    def setUpZope(self, app, configurationContext):
        from App.config import _config
        if not base_hasattr(_config, 'product_config'):
            _config.product_config = {
                'imio.zamqp.core':
                {'ws_url': 'http://localhost:6543', 'ws_password': 'test',
                 'ws_login': 'testuser', 'routing_key': '019999',
                 'client_id': '019999'}}
        super(PMLayer, self).setUpZope(app, configurationContext)

    def setUpPloneSite(self, portal):
        setLocal('request', portal.REQUEST)
        # configure default workflows so Folder has a workflow
        # make sure we have a default workflow
        portal.portal_workflow.setDefaultChain('simple_publication_workflow')
        super(PMLayer, self).setUpPloneSite(portal)
        # Create member area of existing users
        for user in api.user.get_users():
            # this layer is used by imio.pm.wsclient
            _createMemberarea(portal, user.getId())


PM_ZCML = zca.ZCMLSandbox(filename="testing.zcml",
                          package=Products.PloneMeeting,
                          name='PM_ZCML')

PM_Z2 = z2.IntegrationTesting(bases=(z2.STARTUP, PM_ZCML),
                              name='PM_Z2')


PM_TESTING_PROFILE = PMLayer(
    zcml_filename="testing.zcml",
    zcml_package=Products.PloneMeeting,
    additional_z2_products=('Products.PloneMeeting',
                            'collective.eeafaceted.collectionwidget',
                            'Products.CMFPlacefulWorkflow',
                            'Products.PasswordStrength'),
    gs_profile_id='Products.PloneMeeting:testing',
    name="PM_TESTING_PROFILE")

PM_TESTING_PROFILE_INTEGRATION = IntegrationTesting(
    bases=(PM_TESTING_PROFILE,),
    name="PM_TESTING_PROFILE_INTEGRATION")

PM_TESTING_PROFILE_FUNCTIONAL = FunctionalTesting(
    bases=(PM_TESTING_PROFILE,),
    name="PM_TESTING_PROFILE_FUNCTIONAL")

PM_TESTING_ROBOT = FunctionalTesting(
    bases=(
        PM_TESTING_PROFILE,
        REMOTE_LIBRARY_BUNDLE_FIXTURE,
        z2.ZSERVER_FIXTURE,
    ),
    name="PM_TESTING_ROBOT",
)



# -*- coding: utf-8 -*-

from AccessControl import Unauthorized
from appy.gen import No
from datetime import datetime
from datetime import timedelta
from imio.helpers.cache import get_current_user_id
from imio.helpers.security import fplog
from plone import api
from Products.CMFPlone.utils import safe_unicode
from Products.PloneMeeting import logger
from Products.PloneMeeting.external.config import API_URL
from Products.PloneMeeting.external.config import AUTH_CURL_COMMAND
from Products.PloneMeeting.external.config import AUTH_INFOS_ATTR
from zope.globalrequest import getRequest

import json
import requests
import subprocess
import urllib


try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse


def get_auth_token(expire_treshold=60):
    """Get the auth token and store it on the portal.
       Get it again if expired or expires in less than
       given expire_treshold seconds."""
    portal = api.portal.get()
    auth_infos = getattr(portal, AUTH_INFOS_ATTR, {})
    if not auth_infos or auth_infos['expires_in'] < datetime.now():
        logger.info('Getting authentication token')
        start = datetime.now()
        result = subprocess.check_output(
            AUTH_CURL_COMMAND, shell=True)
        logger.info(datetime.now() - start)
        result = json.loads(result)
        if 'access_token' in result:
            auth_infos['access_token'] = result['access_token']
            # store that expires_in is 60 seconds before real expires_in
            # so we may probably execute one last request
            auth_infos['expires_in'] = datetime.now() + \
                timedelta(seconds=result['expires_in'] - expire_treshold)
            setattr(portal, AUTH_INFOS_ATTR, auth_infos)
    # logger.info(auth_infos['access_token'])
    return auth_infos.get('access_token') or result


def send_json_request(
        endpoint,
        extra_parameters={},
        extra_headers={},
        method='GET',
        data={},
        return_as_raw=False,
        show_message=False):
    """Send a json request and returns decoded response."""
    token = get_auth_token()
    if 'error' in token:
        return No("Error getting token: %s (%s)" % (token['error'], token.get('error_description', '')))
    headers = {
        'Accept': 'application/json',
        'Cache-Control': 'no-store',
        'Pragma': 'no-cache',
        'expires': 'Mon, 26 Jul 1997 05:00:00 GMT',
        'Content-Type': 'application/json; charset=UTF-8',
        'Authorization': "Bearer %s" % token,
    }
    headers.update(extra_headers)
    # prepare url to call, manage extra_parameters
    if extra_parameters:
        extra_parameters = "&" + urllib.urlencode(extra_parameters)
    request = getRequest()
    # make an Manager able to pass an arbitrary external_user_id for testing
    user_id = request.get('external_user_id')
    if user_id:
        tool = api.portal.get_tool('portal_plonemeeting')
        if not tool.isManager(realManagers=True):
            raise Unauthorized
    else:
        user_id = get_current_user_id(request)
    url = API_URL % (endpoint, user_id, extra_parameters or "")
    target = urlparse(url)
    url = target.geturl()
    # manage cache per request for 'GET'
    content = None
    extras = 'method={0} url={1}'.format(method, url)
    if method == 'GET':
        cachekey_id = "send_json_request_cachekey"
        cache = getattr(request, cachekey_id, {})
        cachekey = "%s__%s" % (url, method)
        content = cache.get(cachekey, None)
        if content is not None:
            fplog('cached_json_call', extras=extras)
    if content is None:
        fplog('execute_json_call', extras=extras)
        start = datetime.now()
        me = {
            "GET": requests.get,
            "POST": requests.post,
            "DELETE": requests.delete}.get(method)
        response = me(url, headers=headers, json=data)
        content = response.content
        logger.info(datetime.now() - start)
        if response.status_code >= 300:
            logger.warn(content)
            if show_message:
                api.portal.show_message(safe_unicode(content), request=getRequest())
            content = No("Error status: %d (%s)" % (response.status_code, content))
            content.status_code = response.status_code
        # manage cache per request for 'GET'
        if method == 'GET':
            cache[cachekey] = content
            setattr(request, cachekey_id, cache)
    if not return_as_raw and content:
        return json.loads(content)
    else:
        return content



# -*- coding: utf-8 -*-

from AccessControl import Unauthorized
from collective.contact.plonegroup.utils import get_plone_group_id
from imio.actionspanel.interfaces import IContentDeletable
from imio.helpers.cache import get_plone_groups_for_user
from imio.helpers.content import get_user_fullname
from imio.helpers.content import get_vocab_values
from imio.helpers.workflow import get_state_infos
from imio.history.browser.views import EventPreviewView
from imio.history.interfaces import IImioHistory
from imio.history.utils import get_event_by_time
from imio.history.utils import getLastWFAction
from plone import api
from plone.autoform import directives
from plone.autoform.form import AutoExtensibleForm
from plone.dexterity.browser.edit import DefaultEditForm
from plone.dexterity.browser.view import DefaultView
from plone.memoize import ram
from plone.supermodel import model
from Products.CMFCore.permissions import ModifyPortalContent
from Products.CMFCore.utils import _checkPermission
from Products.Five import BrowserView
from Products.PloneMeeting.browser.advicechangedelay import _reinit_advice_delay
from Products.PloneMeeting.config import PMMessageFactory as _
from Products.PloneMeeting.utils import get_event_field_data
from Products.PloneMeeting.utils import is_proposing_group_editor
from Products.PloneMeeting.utils import isPowerObserverForCfg
from z3c.form import form
from zope import schema
from zope.component import getAdapter
from zope.event import notify
from zope.globalrequest import getRequest
from zope.i18n import translate
from zope.lifecycleevent import Attributes
from zope.lifecycleevent import ObjectModifiedEvent

import json


def _delay_icon(memberIsAdviserForGroup, advice_info):
    """In case this is a delay aware advice, return a delay_icon
       if advice is not_given/hidden_during_redaction."""
    if not memberIsAdviserForGroup:
        return 'advice_with_delay_disabled_big.png'
    else:
        delay_status = advice_info['delay_infos']['delay_status']
        if delay_status == 'still_time':
            return 'advice_with_delay_big_green.png'
        elif delay_status == 'still_time_but_alert':
            return 'advice_with_delay_big_orange.png'
        elif delay_status == 'timed_out':
            return 'advice_with_delay_big_red.png'
        else:
            return 'advice_with_delay_big.png'


class AdvicesIcons(BrowserView):
    """
      Advices displayed as icons.
    """

    def __call___cachekey(method, self):
        '''cachekey method for self.__call__.'''
        tool = api.portal.get_tool('portal_plonemeeting')
        # URL to the advice_type can change if server URL changed
        server_url = self.request.get('SERVER_URL', None)
        cfg = tool.getMeetingConfig(self.context)
        # when advices to add, the add advice icon is displayed
        # this takes PowerAdviser into account as well
        has_advices_to_add = self.context.getAdvicesGroupsInfosForUser(
            compute_to_edit=False)
        # confidential advices
        # check confidential advices if not MeetingManager
        isManager = tool.isManager(cfg)
        may_view_confidential_advices = True
        # bypass if no advices
        # but we need nevertheless to compute has_advices_to_add because
        # we may have no advices and no advices and power adviser may add advice
        advices = self.context.adviceIndex.values()
        isPrivacyViewable = True
        if advices and not isManager:
            isPrivacyViewable = self.context.isPrivacyViewable()
            if isPrivacyViewable:
                # if current user is a power observer that would not see the advice
                # we store user plone groups because a power adviser may see a confidential advice
                # if member of the proposingGroup
                confidential_advices = [advice for advice in advices
                                        if advice["isConfidential"] and
                                        not get_plone_group_id(advice["id"], "advisers") in
                                        get_plone_groups_for_user()]
                may_view_confidential_advices = not confidential_advices or \
                    not isPowerObserverForCfg(cfg, power_observer_types=cfg.getAdviceConfidentialFor())
        return (repr(self.context),
                self.context.adviceIndex._p_mtime,
                server_url,
                has_advices_to_add,
                isPrivacyViewable,
                may_view_confidential_advices)

    @ram.cache(__call___cachekey)
    def __call__(self):
        self.advisableGroups = self.context.getAdvicesGroupsInfosForUser(compute_to_edit=False)
        self.advicesToAdd = self.advisableGroups[0]
        self.advicesToEdit = self.advisableGroups[1]
        self.advicesByType = self.context.getAdvicesByType()
        if self.advicesByType or self.advicesToAdd:
            self.tool = api.portal.get_tool('portal_plonemeeting')
            self.cfg = self.tool.getMeetingConfig(self.context)
            self.portal = api.portal.get()
            self.portal_url = self.portal.absolute_url()
            self.userAdviserOrgUids = self.tool.get_orgs_for_user(suffixes=['advisers'])
            self.advice_infos = self.context.getAdviceDataFor(self.context, ordered=True)
            self.every_advice_types = get_vocab_values(self.tool, 'ConfigAdviceTypes')
            if not self.context.adapted().isPrivacyViewable():
                return '<div style="display: inline">&nbsp;-&nbsp;&nbsp;&nbsp;</div>'
        return super(AdvicesIcons, self).__call__()

    def advicesDelayToWarn(self):
        """We will warn if :
           - 'not_given' are in the addable advices;
           - 'hidden_during_redaction' or 'asked_again' are in the editable advices."""

        advicesToWarn = {}

        def _updateAdvicesToWarn(adviceType):
            smaller_delay = 999
            # if we did not found an advice to warn for current user, maybe there is an advice
            # with delay to be given by another group, we show it too
            for org_uid, advice_info in self.advice_infos.items():
                # find smaller delay
                if (advice_info['type'] == adviceType or
                    (adviceType == 'hidden_during_redaction' and
                     advice_info['hidden_during_redaction'] and advice_info['advice_editable']) or
                    (adviceType == 'considered_not_given_hidden_during_redaction' and
                     advice_info['hidden_during_redaction'] and not advice_info['advice_editable'])) and \
                   advice_info['delay'] and \
                   advice_info['delay_infos']['left_delay'] < smaller_delay:
                    if org_uid in self.userAdviserOrgUids:
                        # determinate delay_icon to use
                        advicesToWarn[adviceType] = advice_info, _delay_icon(True, advice_info)
                    # check if we already have a adviceToWarn, if user was adviser
                    # for this group, it is prioritary
                    elif not advicesToWarn.get(adviceType) or \
                            (advicesToWarn.get(adviceType) and not advicesToWarn[adviceType][1] == 0):
                        advicesToWarn[adviceType] = advice_info, _delay_icon(False, advice_info)
                    else:
                        continue
                    smaller_delay = advice_info['delay_infos']['left_delay']

        _updateAdvicesToWarn('not_given')
        _updateAdvicesToWarn('hidden_during_redaction')
        _updateAdvicesToWarn('considered_not_given_hidden_during_redaction')
        _updateAdvicesToWarn('asked_again')

        return advicesToWarn

    def getAddableAdvicePortalTypes(self, advicesToAdd):
        """ """
        res = []
        for adviceToAdd in advicesToAdd:
            advice_portal_type = self.context.adapted()._advicePortalTypeForAdviser(adviceToAdd)
            if advice_portal_type not in res:
                res.append(advice_portal_type)
        return res


class AdvicesIconsInfos(BrowserView):
    """ """

    def __call__(self, adviceType):
        """ """
        self._initAdvicesInfos(adviceType)
        return self.index()

    def _initAdvicesInfos(self, adviceType):
        """ """
        self.tool = api.portal.get_tool('portal_plonemeeting')
        self.cfg = self.tool.getMeetingConfig(self.context)
        self.portal = api.portal.get()
        self.portal_url = self.portal.absolute_url()
        self.advisableGroups = self.context.getAdvicesGroupsInfosForUser(
            compute_to_add=True, compute_power_advisers=False)
        self.advicesToAdd = self.advisableGroups[0]
        self.advicesToEdit = self.advisableGroups[1]
        self.advicesByType = self.context.getAdvicesByType()
        self.adviceType = adviceType
        self.userAdviserOrgUids = self.tool.get_orgs_for_user(suffixes=['advisers'])
        self.itemReviewState = self.context.query_state()
        org_uid = self.context.getProposingGroup()
        # when on an ItemTemplate, there may be no org_uid
        self.userIsInProposingGroup = org_uid and self.tool.user_is_in_org(org_uid=org_uid) or False
        self.isManager = self.tool.isManager(self.cfg)
        self.isRealManager = self.tool.isManager(realManagers=True)
        # edit proposingGroup comment, only compute if item not decided
        # by default editable by Managers only
        self.userIsProposingGroupCommentEditor = False
        self.userMayEditItem = False
        # when on an ItemTemplate, there may be no org_uid
        if org_uid and self.cfg.getEnableAdviceProposingGroupComment():
            self.userIsProposingGroupCommentEditor = self.isRealManager
            self.userMayEditItem = self.isRealManager
            if not self.context.is_decided(self.cfg, self.itemReviewState):
                self.userIsProposingGroupCommentEditor = self.isManager or \
                    is_proposing_group_editor(org_uid, self.cfg)
                self.userMayEditItem = _checkPermission(ModifyPortalContent, self.context)

    def _initAdviceInfos(self, advice):
        """ """
        self.advice_id = advice['id']
        self.memberIsAdviserForGroup = self.advice_id in self.userAdviserOrgUids
        self.adviceIsInherited = self.context.adviceIsInherited(self.advice_id)
        self.mayEdit = not self.adviceIsInherited and \
            ((self.advicesToEdit and self.advice_id in self.advicesToEdit) or
             (self.isRealManager and not self.adviceType == 'not_given'))
        self.adviceHolder = advice.get('adviceHolder', None) or self.context
        self.adviceHolderIsViewable = self.adviceHolder.isViewable()
        self.obj = advice.get('advice_id', None) and \
            self.adviceHolder.get(advice['advice_id'], None)
        self.show_history = self.obj and self.adviceHolderIsViewable and \
            self.obj.restrictedTraverse('@@contenthistory').show_history()

    def showLinkToInherited(self, adviceHolder):
        """ """
        return bool(self.adviceIsInherited and self.context._appendLinkedItem(
            adviceHolder, self.tool, self.cfg, only_viewable=True))

    def mayRemoveInheritedAdvice(self):
        """To remove an inherited advice, must be :
           - MeetingManager and item is not decided;
           - or adviser for p_advice_id group and current item
             in a itemAdviceEditStates review_state."""
        res = False
        if self.adviceIsInherited:
            if self.tool.isManager(realManagers=True) or \
               (self.tool.isManager(self.cfg) and
                    not self.context.is_decided(self.cfg)):
                res = True
            else:
                if self.cfg.getInheritedAdviceRemoveableByAdviser() and \
                   self.advice_id in self.userAdviserOrgUids and \
                   self.itemReviewState in self.cfg.getItemAdviceStatesForOrg(self.advice_id):
                    res = True
        return res

    def mayDelete(self, advice):
        """ """
        return IContentDeletable(advice).mayDelete()

    def mayView(self):
        """ """
        return self.memberIsAdviserForGroup or \
            self.mayEdit or \
            self.adviceType not in ('hidden_during_redaction',
                                    'considered_not_given_hidden_during_redaction')

    def mayChangeDelay(self):
        """ """
        res = False
        if self.context.adviceIndex[self.advice_id]['delay'] and not self.adviceIsInherited:
            view = self.context.unrestrictedTraverse('@@advice-available-delays')
            view._initAttributes(self.advice_id)
            res = view.listSelectableDelays() or \
                view._mayAccessDelayChangesHistory() or view._mayReinitializeDelay()
        return res

    def delay_icon(self, advice_info):
        """Makes it callable in the template."""
        return _delay_icon(self.memberIsAdviserForGroup, advice_info)

    def state_infos(self, advice):
        return get_state_infos(advice)

    def get_adviser_group_ids(self, advice_id):
        """Return list of Plone groups ids having a role in p_advice_id advice WF."""
        advice_portal_type = self.context._advicePortalTypeForAdviser(advice_id)
        suffixes = ["advisers"]
        # for performance reason, if portal_type is the basic "meetingadvice"
        # we only return the "_advisers" suffixed group
        if advice_portal_type != "meetingadvice":
            local_roles = self.portal.portal_types[advice_portal_type].localroles
            # get every suffixes used by localroles
            suffixes = []
            for review_state, infos in local_roles['advice_group'].items():
                for k, v in infos.items():
                    if k not in suffixes:
                        suffixes.append(k)
        return json.dumps(["{0}_{1}".format(advice_id, suffix) for suffix in suffixes])

    def mayEditProposingGroupComment(self):
        """Proposing group may edit comment if able to edit item (not on an inherited advice).
           Advice comment may be changed by proposingGroup when:
           - member is a group editor (not an observer for example);
           - item is editable or advice is addable/editable."""
        res = False
        if self.cfg.getEnableAdviceProposingGroupComment():
            advice_info = self.context.adviceIndex[self.advice_id]
            if not self.adviceIsInherited:
                if self.userIsProposingGroupCommentEditor and \
                    (self.isRealManager or self.userMayEditItem or
                     (advice_info['advice_addable'] or advice_info['advice_editable'])):
                    res = True
        return res

    def mayViewProposingGroupComment(self):
        """May view comment:
           - Proposing group;
           - Asked advice advisers;
           - (Meeting)Managers."""
        res = False
        if self.cfg.getEnableAdviceProposingGroupComment() or \
           self.context.adviceIndex[self.advice_id]["proposing_group_comment"]:
            # bypass for (Meeting)Managers
            if self.isManager or \
               self.memberIsAdviserForGroup or \
               self.userIsInProposingGroup:
                res = True
        return res


class AdviceInfos(BrowserView):
    """Main infos shared between advice popup and advice view."""

    def __call__(self, advice_uid, displayedReviewState, customMessageInfos):
        """ """
        self.portal = api.portal.get()
        self.portal_url = self.portal.absolute_url()
        self.tool = api.portal.get_tool('portal_plonemeeting')
        self.advice = self.context.adviceIndex.get(advice_uid)
        self.adviceType = self.advice['type']
        self.adviceHolder = self.advice.get('adviceHolder', None) or self.context
        self.obj = self.adviceHolder.get(self.advice['advice_id']) if \
            self.advice.get('advice_id', None) else None
        self.displayedReviewState = displayedReviewState
        self.customMessageInfos = customMessageInfos
        self.advice_given_by = self.get_advice_given_by()
        if self.obj is not None:
            # show_history
            adviceHolder = self.advice.get('adviceHolder', None) or self.obj
            adviceHolderIsViewable = adviceHolder.isViewable()
            self.show_history = self.obj and adviceHolderIsViewable and \
                self.obj.restrictedTraverse('@@contenthistory').show_history()
        return self.index()

    def adviser_users(self, advice_info):
        """ """
        res = u''
        if advice_info['userids']:
            res = self.context._displayAdviserUsers(
                advice_info['userids'], self.portal_url, self.tool)
        return res

    def get_advice_given_by(self):
        """The advice was given by advice WF transition before "giveAdvice"."""
        given_by = None
        if self.obj and self.obj._get_final_state_id() != 'advice_given':
            # if final_state_id is actually 'advice_given' we can not compute
            # given_by we are only able to know who created the advice
            last_before_give_advice_event = getLastWFAction(
                self.obj,
                transition=self.obj._get_final_transition_id(),
                ignore_previous_event_actions=['backToAdviceInitialState'])
            if last_before_give_advice_event:
                given_by = get_user_fullname(last_before_give_advice_event["actor"])
        return given_by

    def get_user_fullname(self, user_id):
        return get_user_fullname(user_id)


class ChangeAdviceHiddenDuringRedactionView(BrowserView):
    """View that toggle the advice.advice_hide_during_redaction attribute."""

    def __call__(self):
        # user must be able to edit the advice
        if not _checkPermission(ModifyPortalContent, self.context):
            raise Unauthorized
        else:
            # toggle the value
            self.context.advice_hide_during_redaction = not bool(self.context.advice_hide_during_redaction)
            # when advice_hide_during_redaction, it is handled by ObjectModifiedEvent
            notify(ObjectModifiedEvent(
                self.context,
                Attributes(None, 'advice_hide_during_redaction')))
            if self.request.RESPONSE.status != 200:
                self.request.RESPONSE.status = 200
                if self.request.get('HTTP_REFERER') != self.request.RESPONSE.getHeader('location'):
                    return self.request.RESPONSE.getHeader('location')


class ChangeAdviceAskedAgainView(BrowserView):
    """View that change advice from someting to 'asked_again' and
       from 'asked_again' back to original advice."""

    def __call__(self):
        """ """
        parent = self.context.getParentNode()
        if self.context.advice_type != 'asked_again':
            # we are about to set advice to 'asked_again'
            if not parent.adapted().mayAskAdviceAgain(self.context):
                raise Unauthorized
            # historize the given advice if it was modified since last version
            self.context.historize_if_relevant(comment='advice_asked_again_and_historized_comments')
            # now we may change advice_type to 'asked_again'
            self.context.advice_type = 'asked_again'
            # and we may also set 'advice_hide_during_redaction' to the default
            # XXX for now we do not manage MeetingConfig.defaultAdviceHiddenDuringRedaction
            # when advice is "asked_again", see https://support.imio.be/browse/PM-3883
            # value defined in the MeetingConfig
            # tool = api.portal.get_tool('portal_plonemeeting')
            # cfg = tool.getMeetingConfig(self.context)
            # self.context.advice_hide_during_redaction = \
            #     bool(self.context.portal_type in cfg.getDefaultAdviceHiddenDuringRedaction())
            # reinitialize advice delay if relevant
            advice_uid = self.context.advice_group
            if parent.adviceIndex[advice_uid]['delay']:
                _reinit_advice_delay(parent, advice_uid)
        else:
            # we are about to set the advice back to original value
            if not parent.adapted().mayBackToPreviousAdvice(self.context):
                raise Unauthorized
            # get last version_id and fall back to it
            adapter = getAdapter(self.context, IImioHistory, 'advice_given')
            adapter.revert_to_last_event()

        notify(ObjectModifiedEvent(self.context))
        item_state = parent.query_state()
        parent._sendAdviceToGiveMailIfRelevant(old_review_state=item_state,
                                               new_review_state=item_state,
                                               force_resend_if_in_advice_review_states=True)


class AdviceConfidentialityView(BrowserView):
    """Display advice confidentiality infos."""

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.portal_url = api.portal.get().absolute_url()

    def __call__(self, advice):
        self.advice = advice
        return super(AdviceConfidentialityView, self).__call__()


class AdviceEventPreviewView(EventPreviewView):
    """ """

    def _update(self, event):
        """ """
        self.tool = api.portal.get_tool('portal_plonemeeting')
        self.cfg = self.tool.getMeetingConfig(self.context)
        self.advice_style = self.cfg.getAdviceStyle()
        # store some advice data
        self.advice_url = self.context.absolute_url()
        self.advice_type = get_event_field_data(event["advice_data"], "advice_type")
        self.advice_comment = get_event_field_data(event["advice_data"], "advice_comment")
        self.advice_observations = get_event_field_data(event["advice_data"], "advice_observations")
        self.event_time = repr(float(event['time']))
        self.userAdviserOrgUids = self.tool.get_orgs_for_user(suffixes=['advisers'])
        self.userOrgUids = self.tool.get_orgs_for_user()

    def __call__(self, event):
        self._update(event)
        return super(AdviceEventPreviewView, self).__call__(event)

    def may_view_historized_data(self):
        """Viewable by:
           - (Meeting)Managers
           - proposingGroup members;
           - advice _advisers group members."""
        return self.tool.isManager(self.cfg) or \
            self.context.advice_group in self.userAdviserOrgUids or \
            self.context.aq_parent.getProposingGroup() in self.userOrgUids


class AdviceGivenHistoryView(BrowserView):
    """ """

    def __call__(self, event_time):
        """ """
        event = get_event_by_time(self.context, 'advice_given', float_event_time=event_time)
        view = self.context.restrictedTraverse('@@history-event-preview')
        view._update(event)
        if not view.may_view_historized_data():
            raise Unauthorized
        self.advice_data = event['advice_data']
        self.item_data = event['item_data']
        return super(AdviceGivenHistoryView, self).__call__()


def _display_asked_again_warning(advice, parent):
    """If advice is "asked_again" and current user is adviser
       for advice_group display a message explaining to change the advice_type."""
    if advice.advice_type == "asked_again":
        advisers_group_id = get_plone_group_id(advice.advice_group, 'advisers')
        if parent.adviceIndex[advice.advice_group]['advice_editable']:
            if advisers_group_id in get_plone_groups_for_user():
                api.portal.show_message(
                    _("warning_advice_asked_again_need_to_change_advice_type"),
                    request=advice.REQUEST, type="warning")


class AdviceView(DefaultView):
    """ """

    def __call__(self):
        """Check if viewable by current user in case smart guy call the right url."""
        self.parent = self.context.aq_inner.aq_parent
        self.portal = api.portal.get()
        self.portal_url = self.portal.absolute_url()
        self.advice_icons_infos = self.parent.unrestrictedTraverse('@@advices-icons-infos')
        advice_type = self.parent._shownAdviceTypeFor(self.parent.adviceIndex[self.context.advice_group])
        self.advice_icons_infos._initAdvicesInfos(advice_type)
        self.advice_icons_infos._initAdviceInfos(self.parent.adviceIndex[self.context.advice_group])
        if not self.advice_icons_infos.mayView():
            raise Unauthorized
        _display_asked_again_warning(self.context, self.parent)
        return super(AdviceView, self).__call__()


class AdviceEdit(DefaultEditForm):
    """
        Edit form redefinition to display message when advice "asked_again".
    """

    def update(self):
        super(AdviceEdit, self).update()
        if not self.actions.executedActions:
            _display_asked_again_warning(self.context, self.context.aq_inner.aq_parent)
        # check if need to auto hide the advice if it is "asked_again"
        if self.context.advice_type == 'asked_again' and \
           self.context.advice_hide_during_redaction is False:
            tool = api.portal.get_tool('portal_plonemeeting')
            cfg = tool.getMeetingConfig(self.context)
            if self.context.portal_type in cfg.getDefaultAdviceHiddenDuringRedaction():
                api.portal.show_message(_("advice_hide_during_redaction_set_auto_to_true"),
                                        request=self.request)
                self.widgets['advice_hide_during_redaction'].value = ['true']


def advice_uid_default():
    """
      Get the value from the REQUEST as it is passed when calling the
      form : form?advice_uid=advice_uid.
    """
    request = getRequest()
    return request.get('advice_id', u'')


class IBaseAdviceInfoSchema(model.Schema):

    directives.mode(advice_uid='hidden')
    advice_uid = schema.TextLine(
        title=_(u"Advice uid"),
        description=_(u""),
        defaultFactory=advice_uid_default,
        required=False)


class BaseAdviceInfoForm(AutoExtensibleForm, form.EditForm):
    """
      Base form make to work also when advice is not given.
    """
    label = _(u"")
    description = u''
    schema = IBaseAdviceInfoSchema
    ignoreContext = True  # don't use context to get widget data

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.label = translate(
            self.label, domain='PloneMeeting', context=self.request)

    def _advice_infos(self, data, context=None):
        '''Init @@advices-icons-infos and returns it.'''
        context = context or self.context
        # check if may remove inherited advice
        advice_infos = context.unrestrictedTraverse('@@advices-icons-infos')
        # initialize advice_infos
        advice_data = context.getAdviceDataFor(context, data['advice_uid'])
        advice_infos(context._shownAdviceTypeFor(advice_data))
        advice_infos._initAdviceInfos(advice_data)
        return advice_infos



# -*- coding: utf-8 -*-

from imio.helpers.setup import load_type_from_package
from imio.pyutils.utils import replace_in_list
from Products.PloneMeeting.migrations import logger
from Products.PloneMeeting.migrations import Migrator
from Products.PloneMeeting.setuphandlers import _configureWebspellchecker
from Products.PloneMeeting.setuphandlers import _installWebspellchecker


class Migrate_To_4214(Migrator):

    def _migrateAdviceEditedItemMailEvents(self):
        """Item mail event "adviceEdited" is replaced by "advice_edited__creators"
           and "adviceEditedOwner" is replaced by "advice_edited__owner"."""
        logger.info('Migrating "adviceEdited" in "MeetingConfig.mailItemEvents"...')
        for cfg in self.tool.objectValues('MeetingConfig'):
            mailItemEvents = list(cfg.getMailItemEvents())
            if "adviceEdited" in mailItemEvents:
                mailItemEvents = replace_in_list(
                    mailItemEvents, "adviceEdited", "advice_edited__creators")
                cfg.setMailItemEvents(mailItemEvents)
            if "adviceEditedOwner" in mailItemEvents:
                mailItemEvents = replace_in_list(
                    mailItemEvents, "adviceEditedOwner", "advice_edited__Owner")
                cfg.setMailItemEvents(mailItemEvents)
        logger.info('Done.')

    def _installIMIOWebSpellChecker(self):
        """Configure imio.webspellchecker."""
        logger.info('Install and configure "imio.webspellchecker"...')
        replaced = _installWebspellchecker(self.portal)
        if replaced is False:
            self.warn(
                logger,
                "In _installIMIOWebSpellChecker could not remove 'Scayt' "
                "option from toolbar_Custom!")
        _configureWebspellchecker(self.portal)
        logger.info('Done.')

    def _updatePortalTypesTitle(self):
        """Meeting/MeetingItem portal_types title is now the translated version."""
        logger.info('Updating every Meeting/MeetingItem portal_types title...')
        for cfg in self.tool.objectValues('MeetingConfig'):
            cfg.registerPortalTypes()
        logger.info('Done.')

    def run(self, extra_omitted=[], from_migration_to_4200=False):

        logger.info('Migrating to PloneMeeting 4214...')
        # reload ConfigurablePODTemplate to use every_annex_types_vocabulary for field store_as_annex
        load_type_from_package('ConfigurablePODTemplate', 'Products.PloneMeeting:default')
        self._migrateAdviceEditedItemMailEvents()
        self._updatePortalTypesTitle()
        # not done for now, we will enable it when necessary
        # self._installIMIOWebSpellChecker()
        # add text criterion on "item title only" again as it was not in default
        # dashboard faceted criteria, new MeetingConfigs created manually in between
        # are missing this new criterion
        self.updateFacetedFilters(xml_filename='upgrade_step_4211_add_item_widgets.xml')
        # holidays 2025 were added
        self.updateHolidays()
        logger.info('Migrating to PloneMeeting 4214... Done.')


def migrate(context):
    '''This migration function will:

       1) Reload ConfigurablePODTemplate as store_as_annex field vocabulary changed;
       2) Update values of MeetingConfig.itemMailEvents as format of
          "adviceEdited" values changed;
       3) Update Meeting/MeetingItem portal_types title as we store the real title now;
       4) Update faceted filters;
       5) Not done for now: install and configure "imio.webspellchecker".

    '''
    migrator = Migrate_To_4214(context)
    migrator.run()
    migrator.finish()



# -*- coding: utf-8 -*-
#
# File: testAdvices.py
#
# GNU General Public License (GPL)
#

from collective.iconifiedcategory.browser.tabview import CategorizedContent
from collective.iconifiedcategory.config import get_sort_categorized_tab
from collective.iconifiedcategory.config import set_sort_categorized_tab
from collective.iconifiedcategory.utils import get_categorized_elements
from imio.helpers.cache import cleanRamCacheFor
from Products.CMFPlone.utils import safe_unicode
from Products.PloneMeeting.columns import ItemLinkedMeetingColumn
from Products.PloneMeeting.columns import PMAnnexActionsColumn
from Products.PloneMeeting.columns import PMPrettyLinkColumn
from Products.PloneMeeting.config import AddAnnex
from Products.PloneMeeting.config import AddAnnexDecision
from Products.PloneMeeting.tests.PloneMeetingTestCase import PloneMeetingTestCase
from zope.i18n import translate


class testColumns(PloneMeetingTestCase):
    '''Tests various aspects of advices management.
       Advices are enabled for PloneGov Assembly, not for PloneMeeting Assembly.'''

    def test_pm_ItemPrettyLinkColumnWhenNotPrivacyViewable(self):
        """When item is not privacyViewable :
           - no link is rendred, only the title;
           - more infos are not displayed."""
        cfg = self.meetingConfig
        cfg.setRestrictAccessToSecretItems(True)
        self._setPowerObserverStates(observer_type='restrictedpowerobservers',
                                     states=(self._stateMappingFor('itemcreated'), ))
        self.request.cookies['pmShowDescriptions'] = 'true'

        self.changeUser('pmCreator1')
        # create 2 exactly same items, second will be set 'secret'
        publicItem = self.create('MeetingItem',
                                 title='Public item title',
                                 description='Public item description')
        self.addAnnex(publicItem)
        self.addAnnex(publicItem, relatedTo='item_decision')
        publicItem.setPrivacy('public')
        publicItem._update_after_edit()
        publicBrain = self.catalog(UID=publicItem.UID())[0]
        secretItem = self.create('MeetingItem',
                                 title='Secret item title',
                                 description='Secret item description')
        self.addAnnex(secretItem)
        secretItem.setPrivacy('secret')
        secretItem._update_after_edit()
        secretBrain = self.catalog(UID=secretItem.UID())[0]

        meetingFolder = self.getMeetingFolder()
        faceted_table = meetingFolder.restrictedTraverse('faceted-table-view')
        column = PMPrettyLinkColumn(meetingFolder, self.portal.REQUEST, faceted_table)
        # as a normal user, everything is viewable
        # link to title, more-infos and annexes
        self.assertTrue(publicItem.adapted().isPrivacyViewable())
        self.assertTrue(secretItem.adapted().isPrivacyViewable())
        publicBrainPrettyLinkColumn = column.renderCell(publicBrain)
        secretBrainPrettyLinkColumn = column.renderCell(secretBrain)
        # make sure cache is shared between cell and item view
        self.assertTrue(publicItem.getPrettyLink() in publicBrainPrettyLinkColumn)
        self.assertTrue(secretItem.getPrettyLink() in secretBrainPrettyLinkColumn)
        # link to title
        self.assertTrue("href='{0}'".format(publicBrain.getURL()) in publicBrainPrettyLinkColumn)
        self.assertTrue("href='{0}'".format(secretBrain.getURL()) in secretBrainPrettyLinkColumn)
        # more infos
        self.assertTrue(' class="pmMoreInfo">' in publicBrainPrettyLinkColumn)
        self.assertTrue(' class="pmMoreInfo">' in secretBrainPrettyLinkColumn)
        # annexes
        self.assertTrue(' class="pmMoreInfo">' in publicBrainPrettyLinkColumn)
        self.assertTrue(' class="pmMoreInfo">' in secretBrainPrettyLinkColumn)

        # now as a restricted power observer, secretItem title is only shown (without a link)
        self.changeUser('restrictedpowerobserver1')
        cleanRamCacheFor('Products.PloneMeeting.MeetingItem.isPrivacyViewable')
        self.assertTrue(publicItem.adapted().isPrivacyViewable())
        self.assertFalse(secretItem.adapted().isPrivacyViewable())
        publicBrainPrettyLinkColumn = column.renderCell(publicBrain)
        secretBrainPrettyLinkColumn = column.renderCell(secretBrain)
        # make sure cache is shared between cell and item view
        self.assertTrue(publicItem.getPrettyLink() in publicBrainPrettyLinkColumn)
        self.assertTrue(secretItem.getPrettyLink() in secretBrainPrettyLinkColumn)
        # link to title
        self.assertTrue("href='{0}'".format(publicBrain.getURL()) in publicBrainPrettyLinkColumn)
        # more infos
        self.assertTrue(' class="pmMoreInfo">' in publicBrainPrettyLinkColumn)
        # annexes
        self.assertTrue(' class="pmMoreInfo">' in publicBrainPrettyLinkColumn)
        # the secret item is not accessible
        self.assertEqual(
            secretBrainPrettyLinkColumn,
            u"<div class='pretty_link' title='Secret item title'>"
            u"<span class='pretty_link_icons'>"
            u"<img title='{0}' src='http://nohost/plone/MeetingItem.png' "
            u"style=\"width: 16px; height: 16px;\" /></span>"
            u"<span class='pretty_link_content state-itemcreated'>Secret item title "
            u"<span class='discreet no_access'>(You can not access this element)</span>"
            u"</span></div>".format(self.portal.portal_types[secretItem.portal_type].Title()))

    def test_pm_AnnexActionsColumnShowArrows(self):
        """Arrows are only shown if annex or annexDecision are orderable.
           Only displayed on annexDecisions if only annexDecision addable and no more annex addable."""
        # avoid adding recurring items to created meeting
        self._removeConfigObjectsFor(self.meetingConfig, folders=['recurringitems'])

        self.changeUser('pmManager')
        item = self.create('MeetingItem')
        item.setDecision('<p>Decision text</p>')
        annex1 = self.addAnnex(item)
        annex2 = self.addAnnex(item)

        annex1_infos = get_categorized_elements(item, uids=[annex1.UID()])
        annex1_content = CategorizedContent(item, annex1_infos[0])
        annex2_infos = get_categorized_elements(item, uids=[annex2.UID()])
        annex2_content = CategorizedContent(item, annex2_infos[0])

        meetingFolder = self.getMeetingFolder()
        faceted_table = meetingFolder.restrictedTraverse('faceted-table-view')
        column = PMAnnexActionsColumn(meetingFolder, self.portal.REQUEST, faceted_table)
        renderedColumnAnnex1 = column.renderCell(annex1_content)
        renderedColumnAnnex2 = column.renderCell(annex2_content)
        self.assertTrue(self.hasPermission(AddAnnex, item))
        # sort_categorized_tab must be False to show arrows
        self.assertTrue(get_sort_categorized_tab())
        self.assertFalse('folder_position_typeaware?position=down' in renderedColumnAnnex1)
        self.assertFalse('folder_position_typeaware?position=up' in renderedColumnAnnex2)
        set_sort_categorized_tab(False)
        renderedColumnAnnex1 = column.renderCell(annex1_content)
        renderedColumnAnnex2 = column.renderCell(annex2_content)
        self.assertTrue('folder_position_typeaware?position=down' in renderedColumnAnnex1)
        self.assertTrue('folder_position_typeaware?position=up' in renderedColumnAnnex2)

        # now test when both annex and annexDecision may be added
        self.validateItem(item)
        self.assertTrue(self.hasPermission(AddAnnex, item))
        self.assertTrue(self.hasPermission(AddAnnexDecision, item))
        annexDecision1 = self.addAnnex(item, relatedTo='item_decision')
        annexDecision2 = self.addAnnex(item, relatedTo='item_decision')
        annexDecision1_infos = get_categorized_elements(item, uids=[annexDecision1.UID()])
        annexDecision1_content = CategorizedContent(item, annexDecision1_infos[0])
        annexDecision2_infos = get_categorized_elements(item, uids=[annexDecision2.UID()])
        annexDecision2_content = CategorizedContent(item, annexDecision2_infos[0])

        renderedColumnAnnex1 = column.renderCell(annex1_content)
        renderedColumnAnnex2 = column.renderCell(annex2_content)
        renderedColumnDecisionAnnex1 = column.renderCell(annexDecision1_content)
        renderedColumnDecisionAnnex2 = column.renderCell(annexDecision2_content)
        self.assertTrue('folder_position_typeaware?position=down' in renderedColumnAnnex1)
        self.assertTrue('folder_position_typeaware?position=up' in renderedColumnAnnex2)
        self.assertTrue('folder_position_typeaware?position=down' in renderedColumnDecisionAnnex1)
        self.assertTrue('folder_position_typeaware?position=up' in renderedColumnDecisionAnnex2)
        # and it works
        item.folder_position_typeaware(position='down', id=annex1.getId())
        item.folder_position_typeaware(position='up', id=annex2.getId())
        item.folder_position_typeaware(position='down', id=annexDecision1.getId())
        item.folder_position_typeaware(position='up', id=annexDecision2.getId())

        # now when only annexDecision are addable
        meeting = self.create('Meeting')
        self.presentItem(item)
        self.closeMeeting(meeting)
        self.assertFalse(self.hasPermission(AddAnnex, item))
        self.assertTrue(self.hasPermission(AddAnnexDecision, item))
        renderedColumnAnnex1 = column.renderCell(annex1_content)
        renderedColumnAnnex2 = column.renderCell(annex2_content)
        renderedColumnDecisionAnnex1 = column.renderCell(annexDecision1_content)
        renderedColumnDecisionAnnex2 = column.renderCell(annexDecision2_content)
        self.assertFalse('folder_position_typeaware?position=down' in renderedColumnAnnex1)
        self.assertFalse('folder_position_typeaware?position=up' in renderedColumnAnnex2)
        self.assertTrue('folder_position_typeaware?position=up' in renderedColumnDecisionAnnex1)
        self.assertTrue('folder_position_typeaware?position=down' in renderedColumnDecisionAnnex2)
        # and it works
        item.folder_position_typeaware(position='up', id=annexDecision1.getId())
        item.folder_position_typeaware(position='down', id=annexDecision2.getId())

    def test_pm_ItemLinkedMeetingColumnWhenMeetingNotViewable(self):
        """Test when link to meeting displayed in the items dashboard."""
        self._setPowerObserverStates(states=('presented', ))
        self._setPowerObserverStates(field_name='meeting_states', states=())
        self.changeUser('pmManager')
        item = self.create('MeetingItem')
        self.create('Meeting')
        meetingFolder = self.getMeetingFolder()
        faceted_table = meetingFolder.restrictedTraverse('faceted-table-view')
        column = ItemLinkedMeetingColumn(meetingFolder, self.portal.REQUEST, faceted_table)
        # item not linked to a meeting
        item_brain = self.catalog(UID=item.UID())[0]
        self.assertEqual(column.renderCell(item_brain), u'-')
        self.presentItem(item)

        # linked and viewable
        item_brain = self.catalog(UID=item.UID())[0]
        self.assertTrue(u"<span class='pretty_link_content state-created'>" in column.renderCell(item_brain))
        # linked but not viewable
        self.changeUser('powerobserver1')
        # column have use_caching=True
        column = ItemLinkedMeetingColumn(meetingFolder, self.portal.REQUEST, faceted_table)
        self.assertTrue(u"<span class='pretty_link_content state-created'>" in column.renderCell(item_brain))

    def test_pm_ItemCategoryColumn(self):
        """Test the column displaying category of MeetingItem."""
        self._enableField('category')
        self._enable_column('getCategory')
        self.changeUser('pmCreator1')
        item = self.create('MeetingItem')
        meetingFolder = self.getMeetingFolder()
        faceted_table = meetingFolder.restrictedTraverse('faceted-table-view')
        faceted_table.initColumns()
        column = faceted_table.columnByName['getCategory']
        item_brain = self.catalog(UID=item.UID())[0]
        self.assertEqual(column.renderCell(item_brain), u'Development topics')
        self.assertEqual(
            column.getCSSClasses(item_brain),
            {'td': 'td_cell_getCategory td_cell_development',
             'th': 'th_header_getCategory'})
        item.setCategory('')
        item.reindexObject(idxs=['getCategory'])
        item_brain = self.catalog(UID=item.UID())[0]
        self.assertEqual(column.renderCell(item_brain), u'-')
        self.assertEqual(
            column.getCSSClasses(item_brain),
            {'td': 'td_cell_getCategory td_cell_',
             'th': 'th_header_getCategory'})

    def test_pm_MeetingPrettyLinkColumnWithStaticInfos(self):
        """Test the PMPrettyLinkColumn for Meeting, especially when displaying static infos."""
        cfg = self.meetingConfig
        self._enableField('category')
        self.changeUser('pmManager')
        meeting = self.create('Meeting')
        meetingFolder = self.getMeetingFolder()
        faceted_table = meetingFolder.restrictedTraverse('faceted-table-view')
        faceted_table.initColumns()
        column = faceted_table.columnByName['pretty_link']
        meeting_brain = self.catalog(UID=meeting.UID())[0]
        # header
        faceted_table.batch = []
        self.assertTrue("c0=sortable_title" in column.renderHeadCell())
        self.assertFalse('static-infos' in column.renderCell(meeting_brain))
        # enable static_end_date without enabled in used_attrs
        cfg.setMeetingColumns(('static_end_date', ))
        self.assertFalse('static-infos' in column.renderCell(meeting_brain))
        # now if end_date contains something, it will be displayed
        meeting.end_date = meeting.date
        self.assertTrue('static-infos' in column.renderCell(meeting_brain))

    def test_pm_ReviewStateTitleColumn(self):
        """Will display review_state title by getting the title used
           in the workflow object."""
        cfg = self.meetingConfig
        item_wf = cfg.getItemWorkflow(True)
        self._enable_column('review_state_title')
        self.changeUser('pmCreator1')
        item = self.create('MeetingItem')
        meetingFolder = self.getMeetingFolder()
        faceted_table = meetingFolder.restrictedTraverse('faceted-table-view')
        faceted_table.initColumns()
        column = faceted_table.columnByName['review_state_title']
        item_brain = self.catalog(UID=item.UID())[0]
        self.assertEqual(column.renderCell(item_brain),
                         translate(safe_unicode(item_wf.states['itemcreated'].title),
                                   domain="plone", context=self.request))
        item_wf.states['itemcreated'].title = 'proposed'
        self.assertEqual(column.renderCell(item_brain), u'Proposed')

    def test_pm_ItemNumberColumn(self):
        """Will display the item number on the meeting_view."""
        self.changeUser('pmManager')
        meeting = self.create('Meeting')
        item = meeting.get_items(ordered=True)[0]
        faceted_table = meeting.restrictedTraverse('faceted-table-view')
        faceted_table.initColumns()
        column = faceted_table.columnByName['getItemNumber']
        item_brain = self.catalog(UID=item.UID())[0]
        # moveable for MeetingManagers
        self.assertTrue(u"\u28ff" in column.renderCell(item_brain))
        self.assertTrue(u"data-item_number='100'" in column.renderCell(item_brain))
        self.assertTrue('onclick="moveItem(baseUrl' in column.renderCell(item_brain))
        # header and CSS
        self.assertEqual(
            column.getCSSClasses(item_brain),
            {'td': 'draggable',
             'th': 'th_header_draggable'})
        self.assertEqual(
            column.renderHeadCell(),
            u'</th><th class="th_header_getItemNumber"><script '
            u'type="text/javascript">initializeMeetingItemsDND();</script> ')

        # simplified for non MeetingManagers
        self.changeUser('pmCreator1')
        faceted_table = meeting.restrictedTraverse('faceted-table-view')
        faceted_table.initColumns()
        column = faceted_table.columnByName['getItemNumber']
        self.assertEqual(column.renderCell(item_brain).strip(),
                         u'<span class="itemnumber">1</span>')
        # header and CSS
        self.assertEqual(
            column.getCSSClasses(item_brain),
            {'td': 'td_cell_getItemNumber', 'th': 'th_header_getItemNumber'})
        # no header to avoid init table DND uselessly
        self.assertEqual(column.renderHeadCell(), u' ')

    def test_pm_CopyGroupsColumn(self):
        """Will display the item copyGroups.
           Also because it uses the PMGroups vocabulary."""
        self._enable_column('copyGroups')
        self.changeUser('pmManager')
        item = self.create('MeetingItem', copyGroups=(self.vendors_reviewers, ))
        meetingFolder = self.getMeetingFolder()
        faceted_table = meetingFolder.restrictedTraverse('faceted-table-view')
        faceted_table.initColumns()
        column = faceted_table.columnByName['copyGroups']
        item_brain = self.catalog(UID=item.UID())[0]
        self.assertEqual(column.renderCell(item_brain), u'Vendors (Reviewers)')
        # value is correct without reindex as we use object stored attr
        item.setCopyGroups((self.developers_reviewers, self.vendors_reviewers, ))
        self.assertEqual(column.renderCell(item_brain), u'Developers (Reviewers), Vendors (Reviewers)')


def test_suite():
    from unittest import makeSuite
    from unittest import TestSuite
    suite = TestSuite()
    suite.addTest(makeSuite(testColumns, prefix='test_pm_'))
    return suite



# -*- coding: utf-8 -*-
#
# File: setuphandlers.py
#
# GNU General Public License (GPL)
#

from collections import OrderedDict
from collective.contact.plonegroup.config import PLONEGROUP_ORG
from collective.documentgenerator.config import set_column_modifier
from collective.documentgenerator.config import set_oo_port
from collective.documentgenerator.config import set_oo_server
from collective.documentgenerator.config import set_raiseOnError_for_non_managers
from collective.documentgenerator.config import set_use_stream
from collective.messagesviewlet.utils import add_message
from dexterity.localroles.utils import add_fti_configuration
from eea.facetednavigation.interfaces import ICriteria
from imio.dashboard.setuphandlers import add_orgs_searches
from imio.helpers.catalog import addOrUpdateColumns
from imio.helpers.catalog import addOrUpdateIndexes
from imio.helpers.emailer import get_mail_host
from imio.webspellchecker import config as wsc_config
from plone import api
from Products.CMFPlacefulWorkflow.PlacefulWorkflowTool import WorkflowPolicyConfig_id
from Products.CMFPlone.utils import base_hasattr
from Products.CPUtils.Extensions.utils import configure_ckeditor
from Products.cron4plone.browser.configlets.cron_configuration import ICronConfiguration
from Products.GenericSetup.tool import DEPENDENCY_STRATEGY_REAPPLY
from Products.PloneMeeting.config import CKEDITOR_MENUSTYLES_CUSTOMIZED_MSG
from Products.PloneMeeting.config import HAS_LDAP
from Products.PloneMeeting.config import HAS_SOLR
from Products.PloneMeeting.config import HAS_ZAMQP
from Products.PloneMeeting.config import ManageOwnOrganizationFields
from Products.PloneMeeting.config import PMMessageFactory as _
from Products.PloneMeeting.utils import cleanMemoize
from zope.component import queryUtility
from zope.i18n import translate

import logging
import os


__author__ = """Gaetan DELANNAY <gaetan.delannay@geezteem.com>, Gauthier BASTIEN
<g.bastien@imio.be>, Stephan GEULETTE <s.geulette@imio.be>"""
__docformat__ = 'plaintext'

logger = logging.getLogger('PloneMeeting: setuphandlers')

# Indexes used by PloneMeeting
# XXX warning, do ONLY use ZCTextIndex for real text values,
# NOT returning empty tuple/list like () or [] but empty values like ''
indexInfos = {
    # MeetingItem-related indexes
    'downOrUpWorkflowAgain': ('FieldIndex', {}),
    'getAssociatedGroups': ('KeywordIndex', {}),
    'getCategory': ('FieldIndex', {}),
    'getCompleteness': ('KeywordIndex', {}),
    'getCopyGroups': ('KeywordIndex', {}),
    'getItemNumber': ('FieldIndex', {}),
    'item_is_signed': ('FieldIndex', {}),
    'getGroupsInCharge': ('KeywordIndex', {}),
    'preferred_meeting_uid': ('FieldIndex', {}),
    'preferred_meeting_date': ('DateIndex', {}),
    'getProposingGroup': ('FieldIndex', {}),
    'getRawClassifier': ('FieldIndex', {}),
    'committees_index': ('KeywordIndex', {}),
    'getTakenOverBy': ('FieldIndex', {}),
    'indexAdvisers': ('KeywordIndex', {}),
    'meeting_uid': ('FieldIndex', {}),
    'previous_review_state': ('FieldIndex', {}),
    'reviewProcessInfo': ('FieldIndex', {}),
    'send_to_authority': ('FieldIndex', {}),
    'sentToInfos': ('KeywordIndex', {}),
    'templateUsingGroups': ('KeywordIndex', {}),
    'to_discuss': ('FieldIndex', {}),
    'privacy': ('FieldIndex', {}),
    'pollType': ('FieldIndex', {}),
    'listType': ('FieldIndex', {}),
    'annexes_index': ('KeywordIndex', {}),
    # Meeting-related indexes
    'meeting_date': ('DateIndex', {}),
    # Indexes used by every portal_types
    'getConfigId': ('FieldIndex', {}),
    # a technical index used to store various informations
    'pm_technical_index': ('KeywordIndex', {}),
}
# Metadata to create in portal_catalog
columnInfos = ('getAssociatedGroups',
               'getCategory',
               'meeting_date',
               'committees_index',
               'getGroupsInCharge',
               'preferred_meeting_uid',
               'preferred_meeting_date',
               'getProposingGroup',
               'getRawClassifier',
               'meeting_date',
               'meeting_uid',
               'listType',
               'privacy',
               'pollType',
               'title_or_id')
transformsToDisable = ['word_to_html', 'pdf_to_html', 'pdf_to_text']


def isNotPloneMeetingProfile(context):
    return context.readDataFile("PloneMeeting_marker.txt") is None


def setupHideToolsFromNavigation(context):
    """hide tools"""
    if isNotPloneMeetingProfile(context):
        return
    # uncatalog tools
    site = context.getSite()
    toolnames = ['portal_plonemeeting']
    portal_properties = api.portal.get_tool('portal_properties')
    navtree_properties = getattr(portal_properties, 'navtree_properties')
    if navtree_properties.hasProperty('idsNotToList'):
        for toolname in toolnames:
            try:
                site[toolname].unindexObject()
            except Exception:
                pass
            current = list(navtree_properties.getProperty('idsNotToList') or [])
            if toolname not in current:
                current.append(toolname)
                kwargs = {'idsNotToList': current}
                navtree_properties.manage_changeProperties(**kwargs)


def setupCatalogMultiplex(context):
    """ Configure CatalogMultiplex.

    explicit add classes (meta_types) be indexed in catalogs (white)
    or removed from indexing in a catalog (black)
    """
    if isNotPloneMeetingProfile(context):
        return

    atool = api.portal.get_tool('archetype_tool')
    catalogmap = {}
    catalogmap['ToolPloneMeeting'] = {}
    catalogmap['ToolPloneMeeting']['black'] = ['portal_catalog']
    for meta_type in catalogmap:
        submap = catalogmap[meta_type]
        current_catalogs = set([c.id for c in atool.getCatalogsByType(meta_type)])
        if 'black' in submap:
            for catalog in submap['black']:
                if catalog in current_catalogs:
                    current_catalogs.remove(catalog)
        atool.setCatalogsByType(meta_type, list(current_catalogs))


def postInstall(context):
    """Called at the end of the setup process. """
    if isNotPloneMeetingProfile(context):
        return
    site = context.getSite()
    tool = site.portal_plonemeeting

    activate_solr_and_reindex_if_available(site)
    # Create or update indexes
    addOrUpdateIndexes(site, indexInfos)
    addOrUpdateColumns(site, columnInfos)

    # We add meetingfolder_redirect_view and folder_contents to the list of
    # available views for type "Folder".
    portal_type = site.portal_types.Folder
    available_views = list(portal_type.getAvailableViewMethods(None))
    for folder_view in ('folder_contents', ):
        if folder_view not in available_views:
            available_views.append(folder_view)
    portal_type.manage_changeProperties(view_methods=available_views)

    # Make sure folder "Members" is private
    wft = api.portal.get_tool('portal_workflow')
    if hasattr(site, 'Members') and not wft.getInfoFor(site.Members, 'review_state') == 'private':
        wft.doActionFor(site.Members, 'retract')

    # Make sure portal_plonemeeting permissions are correct regarding used WF
    tool_wf = wft.getWorkflowsFor(tool)[0]
    tool_wf.updateRoleMappingsFor(tool)

    # Make "Unauthorized" exceptions appear in the error log.
    site.error_log.setProperties(
        25, copy_to_zlog=1, ignored_exceptions=('NotFound', 'Redirect', 'Intercepted'))

    # Set a specific workflow policy for all objects created in the tool
    ppw = site.portal_placeful_workflow
    if not hasattr(ppw, 'portal_plonemeeting_policy'):
        ppw.manage_addWorkflowPolicy(
            'portal_plonemeeting_policy',
            workflow_policy_type='default_workflow_policy (Simple Policy)',
            duplicate_id='empty')
        tool.manage_addProduct['CMFPlacefulWorkflow'].manage_addWorkflowPolicyConfig()

    pol = ppw.portal_plonemeeting_policy
    pol.setTitle(_(u'PloneMeeting tool policy'))
    pol.setChain('DashboardCollection', ('',))
    pol.setChainForPortalTypes(
        ('MeetingConfig', ), ('plonemeeting_activity_workflow',))
    # use onestate workflow for Folders contained in the tool/MeetingConfigs
    pol.setChain('Folder', ('plonemeeting_onestate_workflow',))
    pc = getattr(tool, WorkflowPolicyConfig_id)
    pc.setPolicyIn('')
    pc.setPolicyBelow('portal_plonemeeting_policy')

    # We must be able to choose a user password on user creation.
    site.manage_changeProperties(validate_email=0)

    tool.at_post_create_script()

    # Do not generate an action (tab) for each root folder
    site.portal_properties.site_properties.manage_changeProperties(
        disable_folder_sections=True)

    # Display the search box for groups and users
    # when using LDPA or having more that 400 users, we set many_users=True
    many_users = False
    if HAS_LDAP or len(site.acl_users.source_users.listUserIds()) > 400:
        many_users = True
    site.portal_properties.site_properties.manage_changeProperties(
        many_users=many_users)
    site.portal_properties.site_properties.manage_changeProperties(
        many_groups=True)

    # portal_quickinstaller removes some installed elements when reinstalling...
    # re-add them manually here...
    for cfg in tool.objectValues('MeetingConfig'):
        cfg.registerPortalTypes()

    # Check if the personal folder creation of users is enabled
    # check if the creation flag is set to 0, change it if necessary
    if not site.portal_membership.getMemberareaCreationFlag():
        site.portal_membership.setMemberareaCreationFlag()

    # Disable KSS completely
    pjs = site.portal_javascripts
    pjs.unregisterResource('++resource++kukit.js')
    pjs.unregisterResource('++resource++kukit-devel.js')

    # Disable content indexation, we have our own.
    pt = site.portal_transforms
    for transform_id in transformsToDisable:
        if hasattr(pt.aq_base, transform_id):
            pt.manage_delObjects([transform_id])

    # configure CKEditor : adapt available buttons in toolbar and
    # defines it as default Plone editor
    _configureCKeditor(site)

    _configureQuickupload(site)

    _configureWebspellchecker(site)

    # manage safe_html
    _congfigureSafeHtml(site)

    # adapt front-page
    _adaptFrontPage(site)

    # configure imio.pm.zamqp if present
    _configure_zamqp(site)

    # configure collective.documentviewer
    from collective.documentviewer.settings import GlobalSettings
    viewer_settings = GlobalSettings(site)._metadata
    viewer_settings['auto_layout_file_types'] = ['pdf', 'photoshop', 'image',
                                                 'palm', 'ppt', 'txt', 'ps',
                                                 'word', 'rft', 'excel', 'html',
                                                 'visio']
    # do not auto_convert, we will have our own event that will check in portal_plonemeeting
    # if we need to convert the annexes that will be added to an item
    viewer_settings['auto_convert'] = False
    # this avoids layout "documentviewer" being selected for every content_types
    # having a file like AnnexType for example
    viewer_settings['auto_select_layout'] = False
    viewer_settings['pdf_image_format'] = 'jpg'
    viewer_settings['enable_indexation'] = False
    viewer_settings['show_search'] = False
    viewer_settings['show_sidebar'] = False
    viewer_settings['show_search_on_group_view'] = False
    viewer_settings['storage_type'] = 'Blob'

    # configure Products.cron4plone
    # add a call to @@update-delay-aware-advices that will update
    # data regarding the delay-aware advices : call updateAdvices on every items
    # and update the indexAdvisers index in portal_catalog
    cron_configlet = queryUtility(ICronConfiguration, 'cron4plone_config')
    if not cron_configlet.cronjobs:
        # add a cron job that will be launched at 02:00 so set 01:45
        # Syntax: m h dom mon command.
        cron_configlet.cronjobs = [u'45 1 * * portal/@@pm-night-tasks']

    # add a collective.messagesviewlet message that will be used to warn MeetingManagers
    # that there are no more holidays in the configuration in less that 2 months
    messages_config = site.get('messages-config', None)
    if messages_config and 'holidays_warning' not in messages_config.objectIds():
        add_message(id='holidays_warning',
                    title=translate('Holidays warning (do not delete!)',
                                    domain='PloneMeeting',
                                    context=site.REQUEST),
                    text=translate('holidays_warning_message',
                                   domain='PloneMeeting',
                                   context=site.REQUEST),
                    msg_type='significant',
                    req_roles=['Manager', 'MeetingManager'],
                    tal_condition='python: tool.showHolidaysWarning(cfg)',
                    activate=True)
    # if collective.messagesviewlet "browser-warning-ff-chrome" is found, make sure it is enabled
    if messages_config:
        browser_warn_msg = messages_config.get('browser-warning-ff-chrome', None)
        # only enables it if it was never changed, aka created = modified
        if browser_warn_msg and browser_warn_msg.created() == browser_warn_msg.modified() and \
           api.content.get_state(browser_warn_msg) != 'activated':
            api.content.transition(browser_warn_msg, 'activate')

    # collective.documentgenerator : change some default values
    set_oo_server()
    set_oo_port()
    set_column_modifier('nothing')
    set_raiseOnError_for_non_managers(True)
    set_use_stream(False)

    # configure MailHost
    mail_host = get_mail_host()
    mail_host.force_tls = True
    mail_host.smtp_queue = True
    mail_host.smtp_queue_directory = "tmp"
    # (re)start the mail queue
    mail_host._stopQueueProcessorThread()
    mail_host._startQueueProcessorThread()

    # create contacts directory and plonegroup-organization
    if not base_hasattr(site, 'contacts'):
        position_types = [{'name': u'Défaut', 'token': 'default'}, ]
        organization_types = [{'name': u'Défaut', 'token': 'default'}, ]
        organization_levels = [{'name': u'Défaut', 'token': 'default'}, ]
        params = {'title': "Contacts",
                  'position_types': position_types,
                  'organization_types': organization_types,
                  'organization_levels': organization_levels,
                  }
        site.invokeFactory('directory', 'contacts', **params)
        contacts = site['contacts']
        # Organizations creation (in directory)
        params = {'title': u"Mon organisation",
                  'organization_type': u'default', }
        contacts.invokeFactory('organization', PLONEGROUP_ORG, **params)
        own_org = contacts[PLONEGROUP_ORG]
        own_org.manage_permission(
            ManageOwnOrganizationFields, ('Manager', 'Site Administrator'),
            acquire=0)
        # contacts dashboards
        add_orgs_searches(site, add_contact_lists_collections=False)
        # post-configuration
        orgs_searches_folder = contacts.get('orgs-searches')
        # hide the 'review_state' column for orgs related collections
        for org_coll in orgs_searches_folder.objectValues():
            org_coll.customViewFields = [col_name for col_name in org_coll.customViewFields
                                         if col_name != u'review_state']
        # show the "In/out plonegroup organization" filter
        orgs_searches_folder_criteria = ICriteria(orgs_searches_folder)
        own_org_criterion = orgs_searches_folder_criteria.get('c5')
        own_org_criterion.section = 'default'
        own_org_criterion.position = 'center'
        own_org_criterion.hidden = False
        own_org_criterion.default = u'collective.contact.plonegroup.interfaces.IPloneGroupContact'
        # hide active state criterion
        active_org_criterion = orgs_searches_folder_criteria.get('c6')
        active_org_criterion.hidden = True
        # redefine position, need to specify every criteria
        positions = {'top': ['c0', 'c1', 'c4', 'c6', 'c13', 'c14'], 'center': ['c2', 'c3', 'c5']}
        orgs_searches_folder_criteria.position(**positions)
        # clean get_organizations caching
        cleanMemoize(site, prefixes=['plonegroup-utils-get_organizations-'])

    # enable plone.app.caching
    api.portal.set_registry_record('plone.caching.interfaces.ICacheSettings.enabled', True)

    # disable RSS, it does useless catalog search sometimes...
    api.portal.set_registry_record(
        name='Products.CMFPlone.interfaces.syndication.ISiteSyndicationSettings.allowed',
        value=False)
    api.portal.set_registry_record(
        name='Products.CMFPlone.interfaces.syndication.ISiteSyndicationSettings.search_rss_enabled',
        value=False)

    # configure portal_repository
    _configurePortalRepository()

    # configure dexterity localrolesfield
    _configureDexterityLocalRolesField()

    # reorder css
    _reorderCSS(site)


def activate_solr_and_reindex_if_available(site):
    if HAS_SOLR:
        """ activate solr indexing and reindex the existing content """
        from collective.solr.utils import activate
        from collective.solr.utils import getConfig
        if not site.portal_quickinstaller.isProductInstalled('collective.solr'):
            site.portal_setup.runAllImportStepsFromProfile('profile-collective.solr:default')

        solr_activated = api.portal.get_registry_record('collective.solr.active')
        if solr_activated:
            return
        activate(True)
        config = getConfig()
        config.async_indexing = True
        api.portal.set_registry_record('collective.solr.required', [u''])
        port = int(os.environ['SOLR_PORT'])
        api.portal.set_registry_record('collective.solr.port', port)
        import transaction
        transaction.savepoint()
        catalog = api.portal.get_tool('portal_catalog')
        catalog.clearFindAndRebuild()
        transaction.savepoint()
        response = site.REQUEST.RESPONSE
        original = response.write
        response.write = lambda x: x  # temporarily ignore output
        maintenance = site.unrestrictedTraverse("@@solr-maintenance")
        maintenance.clear()
        transaction.savepoint()
        maintenance.reindex()
        response.write = original
        transaction.savepoint()


def _configureCKeditor(site):
    '''Make sure CKeditor is the new default editor used by everyone...
       CKeditor custom styles are kept during migrations using the _before_reinstall/_after_reinstall hooks.'''
    logger.info('Defining CKeditor as the new default editor for every users and configuring it (styles)...')
    # this will install collective.ckeditor if it is not already the case...
    configure_ckeditor(site, custom='plonemeeting', forceTextPaste=0, scayt=0, removeWsc=1)
    # remove every styles defined by default and add the custom styles if not already done...
    cke_props = site.portal_properties.ckeditor_properties
    if cke_props.menuStyles.find(CKEDITOR_MENUSTYLES_CUSTOMIZED_MSG) == -1:
        enc = site.portal_properties.site_properties.getProperty('default_charset')
        msg_highlight_red = translate('ckeditor_style_highlight_in_red',
                                      domain='PloneMeeting',
                                      context=site.REQUEST).encode('utf-8')
        msg_highlight_blue = translate('ckeditor_style_highlight_in_blue',
                                       domain='PloneMeeting',
                                       context=site.REQUEST).encode('utf-8')
        msg_highlight_green = translate('ckeditor_style_highlight_in_green',
                                        domain='PloneMeeting',
                                        context=site.REQUEST).encode('utf-8')
        msg_highlight_yellow = translate('ckeditor_style_highlight_in_yellow',
                                         domain='PloneMeeting',
                                         context=site.REQUEST).encode('utf-8')
        msg_x_small = translate('ckeditor_style_x_small',
                                domain='PloneMeeting',
                                context=site.REQUEST).encode('utf-8')
        msg_small = translate('ckeditor_style_small',
                              domain='PloneMeeting',
                              context=site.REQUEST).encode('utf-8')
        msg_large = translate('ckeditor_style_large',
                              domain='PloneMeeting',
                              context=site.REQUEST).encode('utf-8')
        msg_x_large = translate('ckeditor_style_x_large',
                                domain='PloneMeeting',
                                context=site.REQUEST).encode('utf-8')
        msg_table_optimization = translate('ckeditor_style_table_optimization',
                                           domain='PloneMeeting',
                                           context=site.REQUEST).encode('utf-8')
        msg_table_no_optimization = translate('ckeditor_style_table_no_optimization',
                                              domain='PloneMeeting',
                                              context=site.REQUEST).encode('utf-8')
        msg_indent = translate('ckeditor_style_indent_first_line',
                               domain='PloneMeeting',
                               context=site.REQUEST).encode('utf-8')
        msg_page_break = translate('ckeditor_style_page_break',
                                   domain='PloneMeeting',
                                   context=site.REQUEST).encode('utf-8')

        menu_styles = unicode(
            "[\n{0}\n{{ name : '{1}'\t\t, element : 'span', attributes : {{ 'class' : 'highlight-red' }} }},\n"
            "{{ name : '{2}'\t\t, element : 'span', attributes : {{ 'class' : 'highlight-blue' }} }},\n"
            "{{ name : '{3}'\t\t, element : 'span', attributes : {{ 'class' : 'highlight-green' }} }},\n"
            "{{ name : '{4}'\t\t, element : 'span', attributes : {{ 'class' : 'highlight-yellow' }} }},\n"
            "{{ name : '{5}'\t\t, element : 'p', attributes : {{ 'class' : 'xSmallText' }} }},\n"
            "{{ name : '{6}'\t\t, element : 'p', attributes : {{ 'class' : 'smallText' }} }},\n"
            "{{ name : '{7}'\t\t, element : 'p', attributes : {{ 'class' : 'largeText' }} }},\n"
            "{{ name : '{8}'\t\t, element : 'p', attributes : {{ 'class' : 'xLargeText' }} }},\n"
            "{{ name : '{9}'\t\t, element : 'table', styles : {{ 'table-layout' : 'auto' }} }},\n"
            "{{ name : '{10}'\t\t, element : 'table', styles : {{ 'table-layout' : 'fixed' }} }},\n"
            "{{ name : '{11}'\t\t, element : 'p', attributes : {{ 'style' : 'text-indent: 40px;' }} }},\n"
            "{{ name : '{12}'\t\t, element : 'p', attributes : {{ 'class' : 'page-break' }} }},\n]\n".
            format(CKEDITOR_MENUSTYLES_CUSTOMIZED_MSG,
                   msg_highlight_red,
                   msg_highlight_blue,
                   msg_highlight_green,
                   msg_highlight_yellow,
                   msg_x_small,
                   msg_small,
                   msg_large,
                   msg_x_large,
                   msg_table_optimization,
                   msg_table_no_optimization,
                   msg_indent,
                   msg_page_break), enc)
        cke_props.menuStyles = menu_styles
    # make sure we use resolveuid for images so URL is always correct even if item id changed
    cke_props.allow_link_byuid = True
    # make sure force paste as plain text is disabled
    cke_props.forcePasteAsPlainText = False
    # disable folder creation thru CKeditor to avoid
    # having the add folder icon when adding an image
    cke_props.allow_folder_creation = False
    # set 500px for editor height everywhere
    cke_props.height = '500px'
    # do not use 'rows' of the field widget for editor height
    cke_props.properties_overloaded = (u'width', u'height')
    # use Moono-Lisa skin
    cke_props.skin = u'moono-lisa'
    logger.info('Done.')


def _configureQuickupload(site):
    '''Configure collective.quickupload.'''
    logger.info('Defining collective.quickupload...')
    qu_props = site.portal_properties.quickupload_properties
    qu_props.fill_titles = True
    qu_props.object_unique_id = True
    qu_props.id_as_title = False
    qu_props.sim_upload_limit = 1
    logger.info('Done.')


def installWebspellchecker(context):
    '''XXX temporary custom install for imio.webspellchecker
       to be able to install it only when decided.'''

    if isNotPloneMeetingProfile(context):
        return
    _installWebspellchecker(context.getSite())


def _installWebspellchecker(portal):
    # remove Scayt in CKeditor
    cke_props = portal.portal_properties.ckeditor_properties
    cke_props.enableScaytOnStartup = False
    toolbar_Custom = cke_props.toolbar_Custom
    scayt_values = OrderedDict(
        [
            (",'-','Scayt']", "]"),
            # space after ","
            (", '-','Scayt']", "]"),
            (",'-', 'Scayt']", "]"),
            (", '-', 'Scayt']", "]"),
            # space before "]"
            (", '-','Scayt' ]", "]"),
            (",'-', 'Scayt' ]", "]"),
            (", '-', 'Scayt' ]", "]"),
            # other possibilities, 'Scayt' in the middle
            (",'Scayt',", ","),
            (", 'Scayt',", ",")
        ])
    replaced = False
    for k, v in scayt_values.items():
        if toolbar_Custom.find(k) != -1:
            replaced = True
            toolbar_Custom = toolbar_Custom.replace(k, v)
            cke_props.toolbar_Custom = toolbar_Custom
            break

    # make sure imio.webspellchecker is correctly installed
    # by first uninstalling it because older versions (>1001)
    # profile was not correct
    portal.portal_setup.runAllImportStepsFromProfile(
        'profile-imio.webspellchecker:uninstall')
    portal.portal_setup.runAllImportStepsFromProfile(
        'profile-imio.webspellchecker:default',
        dependency_strategy=DEPENDENCY_STRATEGY_REAPPLY)
    # now that it is installed, configure imio.webspellchecker
    _configureWebspellchecker(portal)
    return replaced


def _configureWebspellchecker(site):
    '''Make sure imio.webspellchecker disallowed_portal_types is correctly configured.'''
    wsc_config.set_enabled(True)
    wsc_config.set_hide_branding(True)
    wsc_config.set_theme('gray')
    portal_types = api.portal.get_tool('portal_types')
    disallowed_portal_types = [pt for pt in portal_types.listContentTypes()
                               if not pt.lower().startswith('meeting') and
                               not pt.startswith('annex') and
                               pt not in ('Message', 'Document', 'News Item', 'Event')]
    wsc_config.set_disallowed_portal_types(disallowed_portal_types)
    wsc_config.set_disable_autosearch_in(
        u'["#form-widgets-title", "#form-widgets-description", '
        u'".select2-focusser", ".select2-input", ".goto_meeting", ".number"]')

def _congfigureSafeHtml(site):
    '''Add some values to safe_html.'''
    logger.info('Adding \'colgroup\' to the list of nasty_tags in safe_html...')
    if u'colgroup' not in site.portal_transforms.safe_html._config['nasty_tags']:
        site.portal_transforms.safe_html._config['nasty_tags'][u'colgroup'] = '1'
    # make sure 'colgroup' and 'col' are not in 'valid_tags'...
    if 'colgroup' in site.portal_transforms.safe_html._config['valid_tags']:
        del site.portal_transforms.safe_html._config['valid_tags']['colgroup']
    if 'col' in site.portal_transforms.safe_html._config['valid_tags']:
        del site.portal_transforms.safe_html._config['valid_tags']['col']
    logger.info('Adding \'strike\' and \'s\' to the list of valid_tags in safe_html...')
    if u'strike' not in site.portal_transforms.safe_html._config['valid_tags']:
        site.portal_transforms.safe_html._config['valid_tags'][u'strike'] = '1'
    if u's' not in site.portal_transforms.safe_html._config['valid_tags']:
        site.portal_transforms.safe_html._config['valid_tags'][u's'] = '1'
    # make sure it was not in 'nasty_tags'...
    if 'strike' in site.portal_transforms.safe_html._config['nasty_tags']:
        del site.portal_transforms.safe_html._config['nasty_tags']['strike']
    if 's' in site.portal_transforms.safe_html._config['nasty_tags']:
        del site.portal_transforms.safe_html._config['nasty_tags']['s']
    # reload transforms so changes are taken into account
    site.portal_transforms.reloadTransforms()


def _adaptFrontPage(site):
    '''
      We adapt the front-page at install time.  To do it only at install time, we will compare front-page modified
      and created attributes, if it is almost equals (in general there is a difference of 0.04 sec, we will
      take here a difference of 0.5 seconds), then it means that it was not changed, we can update it.
    '''
    # first be sure that a front-page exists
    front_page = getattr(site, 'front-page', None)
    if not front_page:
        return

    logger.info('Adapting front-page...')
    # make sure we only adapt it at install time, so when creation and modification date are almost equal
    if front_page.modified() - front_page.created() < 0.000005:
        # disable presentation mode
        front_page.setPresentation(False)
        # there is a difference of less than 0.5 seconds between last modification and creation date
        # it means that practically, the front page was not adapted...
        front_page_title = translate('front_page_title',
                                     domain='PloneMeeting',
                                     context=site.REQUEST)
        front_page.setTitle(front_page_title)
        front_page_descr = translate('front_page_description',
                                     domain='PloneMeeting',
                                     context=site.REQUEST)
        front_page.setDescription(front_page_descr)
        front_page_body = translate('front_page_body',
                                    domain='PloneMeeting',
                                    context=site.REQUEST)
        front_page.setText(front_page_body)
        front_page.reindexObject()
    logger.info('Done.')


def _configure_zamqp(site):
    """Apply imio.zamqp.pm profile if present."""
    if HAS_ZAMQP:
        site.portal_setup.runAllImportStepsFromProfile(
            'imio.zamqp.pm:default',
            dependency_strategy=DEPENDENCY_STRATEGY_REAPPLY)


def _configurePortalRepository(removed_types=[u'ATDocument',
                                              u'ATNewsItem',
                                              u'Document',
                                              u'Event',
                                              u'Link',
                                              u'News Item']):
    """Make sure default Plone content type are not versionable."""
    pr = api.portal.get_tool('portal_repository')
    pr.setVersionableContentTypes([pt_id for pt_id in pr.getVersionableContentTypes()
                                   if pt_id not in removed_types])
    for p_type in removed_types:
        pr._version_policy_mapping.pop(p_type, None)
        # _version_policy_mapping is a dict, make change persistent
        pr._version_policy_mapping = pr._version_policy_mapping


def _configureDexterityLocalRolesField():
    """Configure field meetingadvice.advice_group."""
    # meetingadvice
    roles_config = {'advice_group': {
        'advice_given': {u'advisers': {'rel': '', 'roles': []}},
        'advice_under_edit': {u'advisers': {'rel': '', 'roles': [u'Editor', u'Reviewer', u'Contributor']}}}
    }
    msg = add_fti_configuration(portal_type='meetingadvice',
                                configuration=roles_config['advice_group'],
                                keyname='advice_group',
                                force=True)
    if msg:
        logger.warn(msg)


def _reorderCSS(site):
    """
       Make sure CSS are correctly reordered in portal_css so things
       work as expected...
    """
    portal_css = site.portal_css
    css = ['++resource++collective.eeafaceted.dashboard/collective.eeafaceted.dashboard.css',
           'plonemeeting.css',
           'meetingcommunes.css',
           'imioapps.css',
           'plonemeetingskin.css',
           'imioapps_IEFixes.css',
           'ploneCustom.css']
    for resource in css:
        portal_css.moveResourceToBottom(resource)


def reInstall(context):
    '''Reinstalls the product.'''
    profileId = u'profile-Products.PloneMeeting:default'
    context.runAllImportStepsFromProfile(profileId)



# -*- coding: utf-8 -*-

from imio.helpers.setup import load_type_from_package
from Products.PloneMeeting.migrations import logger
from Products.PloneMeeting.migrations import Migrator


class Migrate_To_4212(Migrator):

    def _enableUseridBehaviorForPerson(self):
        """Field person.userid is now managed by a behavior from
           collective.contact.plonegroup, moreover there is a userid catalog index."""
        logger.info('Enabling "userid" behavior for "person"...')
        load_type_from_package('person', 'Products.PloneMeeting:default')
        self.reindexIndexesFor(idxs=['userid'], **{'portal_type': ['person']})
        logger.info('Done.')

    def run(self, extra_omitted=[], from_migration_to_4200=False):

        logger.info('Migrating to PloneMeeting 4212...')
        if not from_migration_to_4200:
            # this will upgrade collective.contact.plonegroup especially
            self.upgradeAll(omit=['Products.PloneMeeting:default',
                                  self.profile_name.replace('profile-', '')])
            self._enableUseridBehaviorForPerson()
        logger.info('Migrating to PloneMeeting 4212... Done.')


def migrate(context):
    '''This migration function will:

       1) Enable "userid" behavior for "person".
    '''
    migrator = Migrate_To_4212(context)
    migrator.run()
    migrator.finish()
